import datetime as dt
import logging
import os
import re
import warnings
from pathlib import Path
from typing import Any, Union
from urllib.parse import urljoin

import pandas as pd
import xarray as xr
from pandas import DatetimeIndex, Series, Timestamp
from xarray import Dataset

from ravenpy.utilities import gis_import_error_message


try:
    import fiona
except (ImportError, ModuleNotFoundError) as e:
    msg = gis_import_error_message.format(Path(__file__).stem)
    raise ImportError(msg) from e

LOGGER = logging.getLogger("PYWPS")

# Can be set at runtime with `$ env RAVENPY_THREDDS_URL=https://xx.yy.zz/geoserver/ ...`.
THREDDS_URL = os.environ.get("RAVENPY_THREDDS_URL", "https://pavics.ouranos.ca/twitcher/ows/proxy/thredds/")
if not THREDDS_URL.endswith("/"):
    THREDDS_URL = f"{THREDDS_URL}/"

__all__ = [
    "get_CASPAR_dataset",
    "get_ECCC_dataset",
    "get_hindcast_day",
    "get_recent_ECCC_forecast",
    "get_subsetted_forecast",
]


def get_hindcast_day(region_coll: fiona.Collection, date, climate_model="GEPS"):
    """
    Generate a forecast dataset that can be used to run raven.

    Data comes from the CASPAR archive and must be aggregated such that each file
    contains forecast data for a single day, but for all forecast timesteps and
    all members.

    The code takes the region shapefile, the forecast date required, and the
    climate_model to use, here GEPS by default, but eventually could be GEPS, GDPS, REPS or RDPS.
    """
    # Get the file locations and filenames as a function of the climate model and date
    [ds, times] = get_CASPAR_dataset(climate_model, date)

    return get_subsetted_forecast(region_coll, ds, times, True)


def get_CASPAR_dataset(  # noqa: N802
    climate_model: str,
    date: dt.datetime,
    thredds: str = THREDDS_URL,
    directory: str = "dodsC/birdhouse/disk2/caspar/daily/",
) -> tuple[xr.Dataset, list[Union[Union[DatetimeIndex, Series, Timestamp, Timestamp], Any]]]:
    """
    Return CASPAR dataset.

    Parameters
    ----------
    climate_model : str
        Type of climate model, for now only "GEPS" is supported.
    date : dt.datetime
        The date of the forecast.
    thredds : str
        The thredds server url. Default: "https://pavics.ouranos.ca/twitcher/ows/proxy/thredds/".
    directory : str
        The directory on the thredds server where the data is stored. Default: "dodsC/birdhouse/disk2/caspar/daily/".

    Returns
    -------
    xr.Dataset
        The forecast dataset.
    """
    if thredds[-1] != "/":
        warnings.warn("The thredds url should end with a slash. Appending it to the url.", stacklevel=2)
        thredds = f"{thredds}/"

    if climate_model == "GEPS":
        d = dt.datetime.strftime(date, "%Y%m%d")
        file_location = urljoin(directory, f"GEPS_{d}.nc")
        file_url = urljoin(thredds, file_location)
        ds = xr.open_dataset(file_url)
        # Here we also extract the times at 6-hour intervals as Raven must have
        # constant timesteps and GEPS goes to 6 hours
        start = pd.to_datetime(ds.time[0].values)
        times = [start + dt.timedelta(hours=n) for n in range(0, 384, 6)]
    else:
        # Eventually: GDPS, RDPS and REPS
        raise NotImplementedError("Only the GEPS model is currently supported")

    # Checking that these exist.
    for f in ["pr", "tas"]:
        if f not in ds:
            raise AttributeError(f"'{f}' not present in dataset")

    return ds, times


def get_ECCC_dataset(  # noqa: N802
    climate_model: str,
    thredds: str = THREDDS_URL,
    directory: str = "dodsC/datasets/forecasts/eccc_geps/",
) -> tuple[Dataset, list[Union[Union[DatetimeIndex, Series, Timestamp, Timestamp], Any]]]:
    """
    Return latest GEPS forecast dataset.

    Parameters
    ----------
    climate_model : str
        Type of climate model, for now only "GEPS" is supported.
    thredds : str
        The thredds server url. Default: "https://pavics.ouranos.ca/twitcher/ows/proxy/thredds/".
    directory : str
        The directory on the thredds server where the data is stored. Default: "dodsC/datasets/forecasts/eccc_geps/".

    Returns
    -------
    xr.Dataset
        The forecast dataset.
    """
    if thredds[-1] != "/":
        warnings.warn("The thredds url should end with a slash. Appending it to the url.", stacklevel=2)
        thredds = f"{thredds}/"

    if climate_model == "GEPS":
        # Eventually the file will find a permanent home, until then let's use the test folder.
        file_location = urljoin(directory, "GEPS_latest.ncml")
        file_url = urljoin(thredds, file_location)
        ds = xr.open_dataset(file_url)
        # Here we also extract the times at 6-hour intervals as Raven must have
        # constant timesteps and GEPS goes to 6 hours
        start = pd.to_datetime(ds.time[0].values)
        times = [start + dt.timedelta(hours=n) for n in range(0, 384, 6)]
    else:
        # Eventually: GDPS, RDPS and REPS
        raise NotImplementedError("Only the GEPS model is currently supported")

    # Checking that these exist. IF the files are still processing, possible that one or both are not available!
    for f in ["pr", "tas"]:
        if f not in ds:
            raise AttributeError(f"'{f}' not present in dataset")

    return ds, times


def get_recent_ECCC_forecast(  # noqa: N802
    region_coll: fiona.Collection, climate_model: str = "GEPS"
) -> xr.Dataset:
    """
    Generate a forecast dataset that can be used to run raven.

    Data comes from the ECCC datamart and collected daily. It is aggregated
    such that each file contains forecast data for a single day, but for all
    forecast timesteps and all members.

    The code takes the region shapefile and the climate_model to use, here GEPS
    by default, but eventually could be GEPS, GDPS, REPS or RDPS.

    Parameters
    ----------
    region_coll : fiona.Collection
        The region vectors.
    climate_model : str
        Type of climate model, for now only "GEPS" is supported.

    Returns
    -------
    xr.Dataset
        The forecast dataset.
    """
    [ds, times] = get_ECCC_dataset(climate_model)

    # Make the variable name compatible with the hindcasting tools.
    ds = ds.rename({"member": "members"})

    return get_subsetted_forecast(region_coll, ds, times, False)


def get_subsetted_forecast(
    region_coll: fiona.Collection,
    ds: xr.Dataset,
    times: Union[dt.datetime, xr.DataArray],
    is_caspar: bool,
) -> xr.Dataset:
    """
    Get Subsetted Forecast.

    This function takes a dataset, a region and the time sampling array and returns
    the subsetted values for the given region and times.

    Parameters
    ----------
    region_coll : fiona.Collection
        The region vectors.
    ds : xr.Dataset
        The dataset containing the raw, worldwide forecast data.
    times : dt.datetime or xr.DataArray
        The array of times required to do the forecast.
    is_caspar : bool
        True if the data comes from Caspar, false otherwise. Used to define lat/lon on rotated grid.

    Returns
    -------
    xr.Dataset
        The forecast dataset.
    """
    # Extract the bounding box to subset the entire forecast grid to something
    # more manageable
    lon_min = region_coll.bounds[0]
    lon_max = region_coll.bounds[2]
    lat_min = region_coll.bounds[1]
    lat_max = region_coll.bounds[3]

    # Add a very simple lon wraparound if data suggests for it
    if ((ds.lon.min() >= 0) and (ds.lon.max() <= 360)) and (lon_max < 0):
        lon_min += 360
        lon_max += 360

    # Subset the data to the desired location (bounding box) and times
    ds = ds.where(
        (ds.lon <= lon_max) & (ds.lon >= lon_min) & (ds.lat <= lat_max) & (ds.lat >= lat_min),
        drop=True,
    ).sel(time=times)

    # Rioxarray requires CRS definitions for variables
    # Get CRS, e.g. 4326
    crs = int(re.match(r"epsg:(\d+)", region_coll.crs["init"]).group(1))

    # Here the name of the variable could differ based on the Caspar file processing
    tas = ds.tas.rio.write_crs(crs)
    pr = ds.pr.rio.write_crs(crs)
    ds = xr.merge([tas, pr])

    # Now apply the mask of the basin contour and average the values to get a single time series
    if is_caspar:
        ds.rio.set_spatial_dims("rlon", "rlat")
        ds["rlon"] = ds["rlon"] - 360
        # clip the netcdf and average across space.
        shdf = [next(iter(region_coll))["geometry"]]
        forecast = ds.rio.clip(shdf, crs=crs)
        forecast = forecast.mean(dim={"rlat", "rlon"}, keep_attrs=True)

    else:
        ds.rio.set_spatial_dims("lon", "lat")
        ds["lon"] = ds["lon"] - 360
        # clip the netcdf and average across space.
        shdf = [next(iter(region_coll))["geometry"]]
        forecast = ds.rio.clip(shdf, crs=crs)
        forecast = forecast.mean(dim={"lat", "lon"}, keep_attrs=True)

    return forecast



import os
import warnings
from collections import defaultdict
from pathlib import Path
from typing import Union

import pandas

from ravenpy.utilities import gis_import_error_message


try:
    import geopandas
    from osgeo import __version__ as osgeo_version
    from osgeo import ogr, osr
    from shapely import wkt

    osr.UseExceptions()
except (ImportError, ModuleNotFoundError) as e:
    msg = gis_import_error_message.format(Path(__file__).stem)
    raise ImportError(msg) from e

import netCDF4
import numpy as np


def open_shapefile(path: Union[str, os.PathLike]):
    """Return GeoDataFrame from shapefile path."""
    if isinstance(path, (Path, str)):
        if Path(path).suffix == ".zip":
            path = f"zip://{path}"
        df = geopandas.read_file(path)
    elif isinstance(path, geopandas.GeoDataFrame):
        df = path
    else:
        raise NotImplementedError(path)
    return df


class BasinMakerExtractor:
    """
    Encapsulate the logic of converting the Routing Product into the required data structures to generate the RVH file format.

    Parameters
    ----------
    df : GeoDataFrame
        Sub-basin information.
    hru_aspect_convention : {"GRASS", "ArcGIS"}
        How sub-basin aspect is defined.
    routing_product_version : {"2.1", "1.0"}
        Version of the BasinMaker data.
    """

    MAX_RIVER_SLOPE = 0.00001
    USE_LAKE_AS_GAUGE = False
    WEIR_COEFFICIENT = 0.6
    USE_LAND_AS_GAUGE = False
    USE_MANNING_COEFF = False
    MANNING_DEFAULT = 0.035
    HRU_ASPECT_CONVENTION = "GRASS"  # GRASS | ArcGIS
    ROUTING_PRODUCT_VERSION = "2.1"  # 1.0 | 2.1

    def __init__(
        self,
        df,
        hru_aspect_convention=HRU_ASPECT_CONVENTION,
        routing_product_version=ROUTING_PRODUCT_VERSION,
    ):
        self._df = df
        self.hru_aspect_convention = hru_aspect_convention
        self.routing_product_version = routing_product_version

    def extract(self, hru_from_sb: bool = False) -> dict:
        """
        Extract data from the Routing Product shapefile and return dictionaries that can be parsed into Raven Commands.

        Parameters
        ----------
        hru_from_sb : bool
            If True, draw HRU information from subbasin information.
            This is likely to yield crude results.

        Returns
        -------
        dict
            "sub_basins"
               Sequence of dictionaries with `SubBasin` attributes.
            "sub_basin_group"
               Sequence of dictionaries with `SubBasinGroup` attributes.
            "reservoirs"
               Sequence of dictionaries with `Reservoir` attributes.
            "channel_profile"
               Sequence of dictionaries with `ChannelProfile` attributes.
            "hrus"
               Sequence of dictionaries with `HRU` attributes.
        """
        # As one subbasin can be mapped to many rows (HRUs), we use a dict,
        # keyed by their IDs, which we'll transform into a list in the end
        subbasin_recs = {}  #
        land_sb_ids = []
        lake_sb_ids = []
        reservoirs = []  # those are meant to be injected inline in the RVH
        channel_profiles = []  # those are meant to be injected inline in the RVP
        hru_recs = []  # each row corresponds to a HRU

        # Collect all subbasin_ids for fast lookup in next loop
        subbasin_ids = {int(row["SubId"]) for _, row in self._df.iterrows()}

        for _, row in self._df.iterrows():
            # HRU
            if hru_from_sb:
                hru_recs.append(self._extract_hru_from_sb(row))
            else:
                hru_recs.append(self._extract_hru(row))

            subbasin_id = int(row["SubId"])

            is_lake = row["Lake_Cat"] > 0
            if not hru_from_sb:
                is_lake = is_lake and row["HRU_IsLake"] > 0

            if is_lake:
                lake_sb_ids.append(subbasin_id)
                reservoirs.append(self._extract_reservoir(row))
            else:
                land_sb_ids.append(subbasin_id)

            # Subbasin
            sb = self._extract_subbasin(row, is_lake, subbasin_ids)

            if sb["subbasin_id"] not in subbasin_recs:
                subbasin_recs[sb["subbasin_id"]] = sb

                # ChannelProfile
                channel_profiles.append(self._extract_channel_profile(row))

        return dict(
            sub_basins=list(subbasin_recs.values()),
            sub_basin_group=[
                {"name": "Land", "sb_ids": land_sb_ids},
                {"name": "Lakes", "sb_ids": lake_sb_ids},
            ],
            reservoirs=reservoirs,
            channel_profile=channel_profiles,
            hrus=hru_recs,
        )

    def _extract_subbasin(self, row, is_lake, subbasin_ids) -> dict:
        subbasin_id = int(row["SubId"])
        # is_lake = row["HRU_IsLake"] >= 0
        riv_length_field = "Rivlen" if self.routing_product_version == "1.0" else "RivLength"
        # Correctly setting the river length to zero for sub-basins with no channel routing, such as sub-basins with lakes or headwater basins
        if is_lake or row[riv_length_field] <= 1.0:
            river_length_in_kms = 0
        else:
            river_length_in_kms = round(row[riv_length_field] / 1000, 5)
        # river_slope = max(
        #     row["RivSlope"], RoutingProductShapefileExtractor.MAX_RIVER_SLOPE
        # )
        # downstream_id
        downstream_id = int(row["DowSubId"])
        if downstream_id == subbasin_id:
            downstream_id = -1
        elif downstream_id not in subbasin_ids:
            downstream_id = -1
        has_gauge_field = "IsObs" if self.routing_product_version == "1.0" else "Has_Gauge"
        gauged = row[has_gauge_field] > 0 or (is_lake and BasinMakerExtractor.USE_LAKE_AS_GAUGE)
        gauge_id = row["Obs_NM"] if gauged else ""
        return dict(
            subbasin_id=subbasin_id,
            name=f"sub_{subbasin_id}",
            downstream_id=downstream_id,
            profile=f"chn_{subbasin_id}",
            reach_length=river_length_in_kms,
            gauged=gauged,
            gauge_id=gauge_id,
        )

    @staticmethod
    def _extract_reservoir(row) -> dict:
        lake_id = int(row["HyLakeId"])

        return dict(
            subbasin_id=int(row["SubId"]),
            hru_id=int(row.get("HRU_ID", row["SubId"])),
            name=f"Lake_{lake_id}",
            weir_coefficient=BasinMakerExtractor.WEIR_COEFFICIENT,
            crest_width=row["BkfWidth"],
            max_depth=row["LakeDepth"],
            lake_area=row["LakeArea"],
        )

    @staticmethod
    def _extract_channel_profile(row) -> dict:
        subbasin_id = int(row["SubId"])

        # Correctly defining the reach attributes for sub-basins with no channel routing (i.e., lakes or headwater basins)
        channel_elev = row["MeanElev"]
        if row["RivLength"] > 1.0:
            slope = max(row["RivSlope"], BasinMakerExtractor.MAX_RIVER_SLOPE)
            # SWAT: top width of channel when filled with water; bankfull width W_bnkfull
            channel_width = max(row["BkfWidth"], 1)
            # SWAT: depth of water in channel when filled to top of bank
            channel_depth = max(row["BkfDepth"], 1)
            floodn = row["FloodP_n"]
            channeln = row["Ch_n"]
        else:
            slope = 0.12345
            channeln = 0.12345
            floodn = 0.12345
            channel_width = 0.12345
            channel_depth = 0.12345

        # channel profile calculations are based on theory SWAT model is based on
        # see: https://swat.tamu.edu/media/99192/swat2009-theory.pdf
        #      --> "Channel Characteristics" p. 429 ff

        # inverse of channel side slope; channel sides assumed to have 2:1 run to rise ratio
        zch = 2
        # river side width
        sidwd = zch * channel_depth
        # river bottom width W_btm
        botwd = channel_width - 2 * sidwd

        # if derived bottom width is negative, set bottom width to 0.5 * bankfull width and recalculate zch
        # FIXME: zch recalculation is recalculated but not used. Should this be higher up?
        if botwd < 0:
            botwd = 0.5 * channel_width
            sidwd = 0.5 * 0.5 * channel_width
            zch = (channel_width - botwd) / (2 * channel_depth)

        # inverse of floodplain side slope; flood plain side slopes assumed to have 4:1 run to rise ratio
        zfld = 4 + channel_elev
        # floodplain bottom width
        zbot = channel_elev - channel_depth
        # floodplain side width
        sidwdfp = 4 / 0.25

        # geometry of the channel and floodplain
        # (see figure 7:1-2 in SWAT theory document)
        survey_points = (
            (0, zfld),
            (sidwdfp, channel_elev),
            (sidwdfp + 2 * channel_width, channel_elev),
            (sidwdfp + 2 * channel_width + sidwd, zbot),
            (sidwdfp + 2 * channel_width + sidwd + botwd, zbot),
            (sidwdfp + 2 * channel_width + 2 * sidwd + botwd, channel_elev),
            (sidwdfp + 4 * channel_width + 2 * sidwd + botwd, channel_elev),
            (2 * sidwdfp + 4 * channel_width + 2 * sidwd + botwd, zfld),
        )

        if BasinMakerExtractor.USE_MANNING_COEFF:
            mann = channeln
        else:
            mann = BasinMakerExtractor.MANNING_DEFAULT

        # roughness zones of channel and floodplain
        roughness_zones = (
            (0, floodn),
            (sidwdfp + 2 * channel_width, mann),
            (sidwdfp + 2 * channel_width + 2 * sidwd + botwd, floodn),
        )

        return dict(
            name=f"chn_{subbasin_id}",
            bed_slope=slope,
            survey_points=survey_points,
            roughness_zones=roughness_zones,
        )

    def _extract_hru_from_sb(self, row) -> dict:
        """Here we assume one HRU per sub-basin."""
        aspect = row["BasAspect"]

        if self.hru_aspect_convention == "GRASS":
            aspect -= 360
            if aspect < 0:
                aspect += 360
        elif self.hru_aspect_convention == "ArcGIS":
            aspect = 360 - aspect
        else:
            raise ValueError(f"Unknown aspect convention: {self.hru_aspect_convention}")

        return dict(
            hru_id=int(row["SubId"]),
            area=row["BasArea"] / 1e6,
            elevation=row["MeanElev"],
            latitude=row["centroid_y"],
            longitude=row["centroid_x"],
            subbasin_id=int(row["SubId"]),
            slope=row["BasSlope"],
            aspect=aspect,
            hru_type="lake" if row["Lake_Cat"] == 1 else "land",
        )

    def _extract_hru(self, row) -> dict:
        aspect = row["HRU_A_mean"]

        if self.hru_aspect_convention == "GRASS":
            aspect -= 360
            if aspect < 0:
                aspect += 360
        elif self.hru_aspect_convention == "ArcGIS":
            aspect = 360 - aspect
        else:
            raise ValueError(f"Unknown aspect convention: {self.hru_aspect_convention}")

        return dict(
            hru_id=int(row["HRU_ID"]),
            area=row["HRU_Area"] / 1_000_000,
            elevation=row["HRU_E_mean"],
            latitude=row["HRU_CenY"],
            longitude=row["HRU_CenX"],
            subbasin_id=int(row["SubId"]),
            aquifer_profile="[NONE]",
            terrain_class="[NONE]",
            land_use_class=row["LAND_USE_C"],
            veg_class=row["VEG_C"],
            soil_profile=row["SOIL_PROF"],
            slope=row["HRU_S_mean"],
            aspect=aspect,
        )


class GridWeightExtractor:
    """
    Class to extract grid weights.

    Notes
    -----
    To use this on HRU GeoJONS created from the Canadian River and Lake Hydrofabric database, set `routing_id_field`
    to `__INDEX__`.

    The original version of this algorithm can be found at: https://github.com/julemai/GridWeightsGenerator
    """

    DIM_NAMES = ("lon_dim", "lat_dim")
    VAR_NAMES = ("longitude", "latitude")
    ROUTING_ID_FIELD = "SubId"
    NETCDF_INPUT_FIELD = "NetCDF_col"
    AREA_ERROR_THRESHOLD = 0.05
    CRS_LLDEG = 4326  # EPSG id of lat/lon (deg) coordinate reference system (CRS)
    CRS_CAEA = 3573  # EPSG id of equal-area coordinate reference system (CRS)

    def __init__(
        self,
        input_file_path,
        routing_file_path,
        dim_names=DIM_NAMES,
        var_names=VAR_NAMES,
        routing_id_field=ROUTING_ID_FIELD,
        netcdf_input_field=NETCDF_INPUT_FIELD,
        gauge_ids=None,
        sub_ids=None,
        area_error_threshold=AREA_ERROR_THRESHOLD,
    ):
        """
        Initialization.

        Parameters
        ----------
        input_file_path : Path or str
            NetCDF file or shapefile with the data to be weighted.
        routing_file_path : Path or str
            Sub-basin delineation.
        dim_names : tuple
            Names of the dimensions in the NetCDF file.
        var_names : tuple
            Names of the variables in the NetCDF file.

        """
        self._dim_names = tuple(dim_names)
        self._var_names = tuple(var_names)
        self._routing_id_field = routing_id_field
        self._netcdf_input_field = netcdf_input_field
        self._gauge_ids = gauge_ids or []
        self._sub_ids = sub_ids or []
        self._area_error_threshold = area_error_threshold

        if self._gauge_ids and self._sub_ids:
            raise ValueError("Only one of gauge_ids or sub_ids can be specified")

        input_file_path = Path(input_file_path)
        if input_file_path.suffix == ".nc":
            # Note that we cannot use xarray because it complains about variables and dimensions
            # having the same name.
            self._input_data = netCDF4.Dataset(input_file_path)
            self._input_is_netcdf = True
        elif input_file_path.suffix in [".zip", ".shp"]:
            self._input_data = open_shapefile(input_file_path)
            self._input_is_netcdf = False
        else:
            raise ValueError("The input file must be a shapefile (.shp or .zip) or NetCDF")

        self._routing_data = open_shapefile(routing_file_path)

        # Add `__INDEX__` column, which is just the index of the dataframe. This is used as the HRUID in
        # CLRH datasets.
        self._routing_data["__INDEX__"] = range(1, len(self._routing_data) + 1)

    def extract(self) -> dict:
        """Return dictionary to create a GridWeights command."""
        self._prepare_input_data()

        # Read routing data

        # WGS 84 / North Pole LAEA Canada
        self._routing_data = self._routing_data.to_crs(epsg=GridWeightExtractor.CRS_CAEA)

        def keep_only_valid_downsubid_and_obs_nm(g):
            """
            Find only the valid DownSubId and Obs_NM values in a group of rows.

            Notes
            -----
            This function receives a group (g) of routing rows that have been grouped by HRU_ID.
            If there is only one row, there is nothing to do (we return it). If there are more than
            one rows, we want to return the one with the DowSubId != -1 (if there are more than one
            we emit a warning). We also want to search for an Obs_NM value different from -9999 in
            any row of the group, and if we find it we set it in the corresponding field of the
            returning row (if we can't find one we also emit a warning).
            """
            if len(g) == 1:
                return g
            rid = self._routing_id_field
            row = g[g["DowSubId"] != -1].copy()
            if len(row) > 1:
                row = row[:1].copy()
                warnings.warn(f"More than one row with ID={row[rid]} having DowSubId = -1", stacklevel=2)
            obs_nm = g[g["Obs_NM"] != -9999]
            if not obs_nm.empty:
                row["Obs_NM"] = obs_nm["Obs_NM"].iloc[0]
            else:
                warnings.warn(f"All values of Obs_NM are -9999 for rows with ID={row[rid]}", stacklevel=2)

            return row

        # Remove duplicate HRU_IDs while making sure that we keep relevant DowSubId and Obs_NM values
        # FIXME: DeprecationWarning: DataFrameGroupBy.apply operated on the grouping columns.
        #  This behavior is deprecated, and in a future version of pandas the grouping columns will be
        #  excluded from the operation. Either pass `include_groups=False` to exclude the groupings or
        #  explicitly select the grouping columns after groupby to silence this warning.
        self._routing_data = self._routing_data.groupby(self._routing_id_field, group_keys=False).apply(
            keep_only_valid_downsubid_and_obs_nm, include_groups=True
        )

        # Make sure those are ints
        self._routing_data.SubId = self._routing_data.SubId.astype(int)
        self._routing_data.DowSubId = self._routing_data.DowSubId.astype(int)

        if self._gauge_ids:
            # Extract the SubIDs of the gauges that were specified at input
            self._sub_ids = self._routing_data.loc[self._routing_data.Obs_NM.isin(self._gauge_ids)].SubId.unique().tolist()
            if not self._sub_ids:
                raise ValueError(f"No shapes were found with gauge ID (Obs_NM) in {self._gauge_ids}")

        if self._sub_ids:
            # Here we want to extract the network of connected subbasins by going upstream via their DowSubId,
            # starting from the list supplied by the user (either directly, or via their gauge IDs).. We first
            # build a map of downSubID -> subID for efficient lookup
            downsubid_to_subids = defaultdict(set)
            for _, r in self._routing_data.iterrows():
                downsubid_to_subids[r.DowSubId].add(r.SubId)

            expanded_sub_ids = set(self._sub_ids)
            prev = expanded_sub_ids.copy()
            while True:
                curr = set()
                for did in prev:
                    curr |= downsubid_to_subids[did]
                if curr:
                    expanded_sub_ids |= curr
                    prev = curr
                else:
                    break

            self._sub_ids = expanded_sub_ids

        # Reduce the initial dataset with the target Sub IDs
        if self._sub_ids:
            self._routing_data = self._routing_data[self._routing_data.SubId.isin(self._sub_ids)]

        # -------------------------------
        # construct all grid cell polygons
        # -------------------------------

        grid_cell_geom_gpd_wkt = self._compute_grid_cell_polygons()

        # -------------------------------
        # Derive overlay and calculate weights
        # -------------------------------

        grid_weights = []

        for _, row in self._routing_data.iterrows():
            poly = ogr.CreateGeometryFromWkt(wkt.dumps(row.geometry))

            area_basin = poly.Area()
            # bounding box around basin (for easy check of proximity)
            enve_basin = poly.GetEnvelope()

            area_all = 0.0
            # ncells = 0

            row_grid_weights = []

            for ilat in range(self._nlat):
                for ilon in range(self._nlon):
                    # bounding box around grid-cell (for easy check of proximity)
                    enve_gridcell = grid_cell_geom_gpd_wkt[ilat][ilon].GetEnvelope()

                    grid_is_close = self._check_proximity_of_envelops(enve_gridcell, enve_basin)

                    # this check decreases runtime DRASTICALLY (from ~6h to ~1min)
                    if not grid_is_close:
                        continue

                    # grid_cell_area = grid_cell_geom_gpd_wkt[ilat][ilon].Area()

                    # "fake" buffer to avoid invalid polygons and weirdos dumped by ArcGIS
                    inter = grid_cell_geom_gpd_wkt[ilat][ilon].Intersection(poly.Buffer(0.0))

                    area_intersect = inter.Area()

                    area_all += area_intersect

                    if area_intersect > 0:
                        hru_id = int(row[self._routing_id_field])
                        cell_id = ilat * self._nlon + ilon
                        weight = area_intersect / area_basin
                        row_grid_weights.append((hru_id, cell_id, weight))

            # mismatch between area of subbasin (routing product) and sum of all contributions of grid cells (model output)
            error = (area_basin - area_all) / area_basin

            if abs(error) > self._area_error_threshold and area_basin > 500000.0:
                # record all basins with errors larger 5% (if basin is larger than 0.5 km2)
                # error_dict[int(ibasin[key_colname])] = [error, area_basin]
                grid_weights += row_grid_weights

            else:
                # adjust such that weights sum up to 1.0
                for hru_id, cell_id, weight in row_grid_weights:
                    corrected_weight = weight * 1.0 / (1.0 - error)
                    grid_weights.append((hru_id, cell_id, corrected_weight))

                # if error < 1.0:
                #     area_all *= 1.0 / (1.0 - error)
                # error = 0.0

        return dict(
            number_hrus=len(self._routing_data),
            number_grid_cells=self._nlon * self._nlat,
            data=tuple(grid_weights),
        )

    def _prepare_input_data(self):
        if self._input_is_netcdf:
            # Raven numbering is:
            #
            #      [      1      2      3   ...     1*nlon
            #        nlon+1 nlon+2 nlon+3   ...     2*nlon
            #           ...    ...    ...   ...     ...
            #           ...    ...    ...   ...  nlat*nlon ]
            #
            # --> Making sure shape of lat/lon fields is like that
            #

            # TODO: Replace with cf_xarray logic
            lon_var = self._input_data.variables[self._var_names[0]]
            lon_dims = lon_var.dimensions
            lon_var = lon_var[:]

            lat_var = self._input_data.variables[self._var_names[1]]
            lat_dims = lat_var.dimensions
            lat_var = lat_var[:]

            if len(lon_dims) == 2 and len(lat_dims) == 2:
                if lon_dims == self._dim_names:
                    lon_var = np.transpose(lon_var)

                if lat_dims == self._dim_names:
                    lat_var = np.transpose(lat_var)

            elif len(lon_dims) == 1 and len(lat_dims) == 1:
                # Coord vars are 1d, make them 2d
                lon_var_size = lon_var.size
                lon_var = np.tile(lon_var, (lat_var.size, 1))
                lat_var = np.transpose(np.tile(lat_var, (lon_var_size, 1)))
                # FIXME: assertions should not be found outside of testing code. Replace with conditional logic.
                assert lon_var.shape == lat_var.shape  # noqa: S101

            else:
                raise ValueError("The coord variables of the input data must have the same number of dimensions (either 1 or 2)")

            self._lath, self._lonh = self._create_gridcells_from_centers(lat_var, lon_var)

            self._nlon = np.shape(lon_var)[1]
            self._nlat = np.shape(lat_var)[0]

        else:
            # input data is a shapefile

            self._input_data = self._input_data.to_crs(epsg=GridWeightExtractor.CRS_CAEA)

            self._nlon = 1  # only for consistency

            # number of shapes in model "discretization" shapefile (not routing toolbox shapefile)
            self._nlat = self._input_data.geometry.count()  # only for consistency

    def _compute_grid_cell_polygons(self):
        grid_cell_geom_gpd_wkt: list[list[list[ogr.Geometry]]] = [[[] for _ilon in range(self._nlon)] for _ilat in range(self._nlat)]

        if self._input_is_netcdf:
            lath = self._lath
            lonh = self._lonh

            for ilat in range(self._nlat):
                for ilon in range(self._nlon):
                    # -------------------------
                    # EPSG:3035   needs a swap before and after transform ...
                    # -------------------------
                    # gridcell_edges = [ [lath[ilat,ilon]    , lonh[ilat,  ilon]    ],   # for some reason need to switch lat/lon that transform works
                    #                    [lath[ilat+1,ilon]  , lonh[ilat+1,ilon]    ],
                    #                    [lath[ilat+1,ilon+1], lonh[ilat+1,ilon+1]  ],
                    #                    [lath[ilat,ilon+1]  , lonh[ilat,  ilon+1]  ]]

                    # tmp = self._routing_data_to_geometry(gridcell_edges, epsg=crs_caea)
                    # tmp.SwapXY()              # switch lat/lon back
                    # grid_cell_geom_gpd_wkt[ilat][ilon] = tmp

                    # -------------------------
                    # EPSG:3573   does not need a swap after transform ... and is much faster than transform with EPSG:3035
                    # -------------------------
                    #
                    # Windows            Python 3.8.5 GDAL 3.1.3 --> lat/lon (Ming)
                    # MacOS 10.15.6      Python 3.8.5 GDAL 3.1.3 --> lat/lon (Julie)
                    # Graham             Python 3.8.2 GDAL 3.0.4 --> lat/lon (Julie)
                    # Graham             Python 3.6.3 GDAL 2.2.1 --> lon/lat (Julie)
                    # Ubuntu 18.04.2 LTS Python 3.6.8 GDAL 2.2.3 --> lon/lat (Etienne)
                    #
                    if osgeo_version < "3.0":
                        gridcell_edges = [
                            [
                                lonh[ilat, ilon],
                                lath[ilat, ilon],
                            ],  # for some reason need to switch lat/lon that transform works
                            [lonh[ilat + 1, ilon], lath[ilat + 1, ilon]],
                            [lonh[ilat + 1, ilon + 1], lath[ilat + 1, ilon + 1]],
                            [lonh[ilat, ilon + 1], lath[ilat, ilon + 1]],
                        ]
                    else:
                        gridcell_edges = [
                            [
                                lath[ilat, ilon],
                                lonh[ilat, ilon],
                            ],  # for some reason lat/lon order works
                            [lath[ilat + 1, ilon], lonh[ilat + 1, ilon]],
                            [lath[ilat + 1, ilon + 1], lonh[ilat + 1, ilon + 1]],
                            [lath[ilat, ilon + 1], lonh[ilat, ilon + 1]],
                        ]

                    tmp = self._shape_to_geometry(gridcell_edges, epsg=GridWeightExtractor.CRS_CAEA)
                    grid_cell_geom_gpd_wkt[ilat][ilon] = tmp

        else:
            for ishape in range(self._nlat):
                idx = np.where(self._input_data[self._netcdf_input_field] == ishape)[0]
                if len(idx) == 0:
                    # print(
                    #     "Polygon ID = {} not found in '{}'. Numbering of shapefile attribute '{}' needs to be [0 ... {}-1].".format(
                    #         ishape, input_file, key_colname_model, nshapes
                    #     )
                    # )
                    raise ValueError("Polygon ID not found.")
                if len(idx) > 1:
                    # print(
                    #     "Polygon ID = {} found multiple times in '{}' but needs to be unique."
                    #     "Numbering of shapefile attribute '{}' needs to be [0 ... {}-1].".format(
                    #         ishape, input_file, key_colname_model, nshapes
                    #     )
                    # )
                    raise ValueError("Polygon ID not unique.")
                idx = idx[0]
                poly = self._input_data.loc[idx].geometry
                grid_cell_geom_gpd_wkt[ishape][0] = ogr.CreateGeometryFromWkt(wkt.dumps(poly)).Buffer(
                    0.0
                )  # We add an empty buffer here to fix problems with bad polygon topology (actually caused by ESRI's historical incompetence)

        return grid_cell_geom_gpd_wkt

    @staticmethod
    def _create_gridcells_from_centers(lat, lon):
        # create array of edges where (x,y) are always center cells
        nlon = np.shape(lon)[1]
        nlat = np.shape(lat)[0]
        lonh = np.empty((nlat + 1, nlon + 1), dtype=np.float64)
        lath = np.empty((nlat + 1, nlon + 1), dtype=np.float64)
        tmp1 = [
            [(lat[ii + 1, jj + 1] - lat[ii, jj]) / 2 for jj in range(nlon - 1)] + [(lat[ii + 1, nlon - 1] - lat[ii, nlon - 2]) / 2]
            for ii in range(nlat - 1)
        ]
        tmp2 = [
            [(lon[ii + 1, jj + 1] - lon[ii, jj]) / 2 for jj in range(nlon - 1)] + [(lon[ii + 1, nlon - 1] - lon[ii, nlon - 2]) / 2]
            for ii in range(nlat - 1)
        ]
        dlat = np.array(tmp1 + [tmp1[-1]])
        dlon = np.array(tmp2 + [tmp2[-1]])
        lonh[0:nlat, 0:nlon] = lon - dlon
        lath[0:nlat, 0:nlon] = lat - dlat

        # make lat and lon one column and row wider such that all
        lonh[nlat, 0:nlon] = lonh[nlat - 1, 0:nlon] + (lonh[nlat - 1, 0:nlon] - lonh[nlat - 2, 0:nlon])
        lath[nlat, 0:nlon] = lath[nlat - 1, 0:nlon] + (lath[nlat - 1, 0:nlon] - lath[nlat - 2, 0:nlon])
        lonh[0:nlat, nlon] = lonh[0:nlat, nlon - 1] + (lonh[0:nlat, nlon - 1] - lonh[0:nlat, nlon - 2])
        lath[0:nlat, nlon] = lath[0:nlat, nlon - 1] + (lath[0:nlat, nlon - 1] - lath[0:nlat, nlon - 2])
        lonh[nlat, nlon] = lonh[nlat - 1, nlon - 1] + (lonh[nlat - 1, nlon - 1] - lonh[nlat - 2, nlon - 2])
        lath[nlat, nlon] = lath[nlat - 1, nlon - 1] + (lath[nlat - 1, nlon - 1] - lath[nlat - 2, nlon - 2])

        return [lath, lonh]

    @staticmethod
    def _shape_to_geometry(shape_from_jsonfile, epsg=None):
        # converts shape read from shapefile to geometry
        # epsg :: integer EPSG code

        ring_shape = ogr.Geometry(ogr.wkbLinearRing)

        for ii in shape_from_jsonfile:
            ring_shape.AddPoint_2D(ii[0], ii[1])
        # close ring
        ring_shape.AddPoint_2D(shape_from_jsonfile[0][0], shape_from_jsonfile[0][1])

        poly_shape = ogr.Geometry(ogr.wkbPolygon)
        poly_shape.AddGeometry(ring_shape)

        if epsg:
            source = osr.SpatialReference()
            # usual lat/lon projection
            source.ImportFromEPSG(GridWeightExtractor.CRS_LLDEG)

            target = osr.SpatialReference()
            target.ImportFromEPSG(epsg)  # any projection to convert to

            transform = osr.CoordinateTransformation(source, target)
            poly_shape.Transform(transform)

        return poly_shape

    @staticmethod
    def _check_proximity_of_envelops(gridcell_envelop, shape_envelop):
        # checks if two envelops are in proximity (intersect)

        # minX  --> env[0]
        # maxX  --> env[1]
        # minY  --> env[2]
        # maxY  --> env[3]

        return (
            gridcell_envelop[0] <= shape_envelop[1]
            and gridcell_envelop[1] >= shape_envelop[0]
            and gridcell_envelop[2] <= shape_envelop[3]
            and gridcell_envelop[3] >= shape_envelop[2]
        )

    @staticmethod
    def _check_gridcell_in_proximity_of_shape(gridcell_edges, shape_from_jsonfile):
        # checks if a grid cell falls into the bounding box of the shape
        # does not mean it intersects but it is a quick and cheap way to
        # determine cells that might intersect

        # gridcell_edges = [(lon1,lat1),(lon2,lat2),(lon3,lat3),(lon4,lat4)]
        # shape_from_jsonfile

        min_lat_cell = np.min([ii[1] for ii in gridcell_edges])
        max_lat_cell = np.max([ii[1] for ii in gridcell_edges])
        min_lon_cell = np.min([ii[0] for ii in gridcell_edges])
        max_lon_cell = np.max([ii[0] for ii in gridcell_edges])

        lat_shape = np.array([icoord[1] for icoord in shape_from_jsonfile])  # is it lat???
        lon_shape = np.array([icoord[0] for icoord in shape_from_jsonfile])  # is it lon???

        min_lat_shape = np.min(lat_shape)
        max_lat_shape = np.max(lat_shape)
        min_lon_shape = np.min(lon_shape)
        max_lon_shape = np.max(lon_shape)

        return min_lat_cell <= max_lat_shape and max_lat_cell >= min_lat_shape and min_lon_cell <= max_lon_shape and max_lon_cell >= min_lon_shape


def upstream_from_id(
    fid: Union[str, int, float], df: Union[pandas.DataFrame, geopandas.GeoDataFrame]
) -> Union[pandas.DataFrame, geopandas.GeoDataFrame]:
    """
    Return upstream sub-basins by evaluating the downstream networks.

    Parameters
    ----------
    fid : str or int or float
        feature ID of the downstream feature of interest.
    df : pandas.DataFrame or geopandas.GeoDataFrame
        A GeoDataframe comprising the watershed attributes.

    Returns
    -------
    pandas.DataFrame or geopandas.GeoDataFrame
        Basins ids including `fid` and its upstream contributors.
    """
    from ravenpy.utilities.geo import determine_upstream_ids

    return determine_upstream_ids(fid, df, basin_field="SubId", downstream_field="DowSubId")


def upstream_from_coords(
    lon: float, lat: float, df: Union[pandas.DataFrame, geopandas.GeoDataFrame]
) -> Union[pandas.DataFrame, geopandas.GeoDataFrame]:
    """
    Return the sub-basins located upstream from outlet.

    Parameters
    ----------
    lon : float
        Longitude of outlet.
    lat : float
        Latitude of outlet.
    df : pandas.DataFrame or geopandas.GeoDataFrame
        Routing product.

    Returns
    -------
    pandas.DataFrame or geopandas.GeoDataFrame
        Sub-basins located upstream from outlet.
    """
    from ravenpy.utilities.geo import find_geometry_from_coord

    # Find the outlet sub-basin ID
    out_sb = find_geometry_from_coord(lon, lat, df)
    out_sb_id = int(out_sb["SubId"].iloc[0])

    # Find upstream sub-basins
    up_ids = upstream_from_id(out_sb_id, df)

    return up_ids



from pathlib import Path

import click
import numpy as np

from ravenpy.utilities import gis_import_error_message


try:
    import geopandas as gpd
except (ImportError, ModuleNotFoundError) as e:
    msg = gis_import_error_message.format(Path(__file__).stem)
    raise ImportError(msg) from e


COLUMN_NAMES_CONSTANT_HRU = [
    "SubId",
    "DowSubId",
    "RivSlope",
    "RivLength",
    "BasSlope",
    "BasAspect",
    "BasArea",
    "BkfWidth",
    "BkfDepth",
    "Lake_Cat",
    "HyLakeId",
    "LakeVol",
    "LakeDepth",
    "LakeArea",
    "Laketype",
    "Has_Gauge",
    "MeanElev",
    "FloodP_n",
    "Q_Mean",
    "Ch_n",
    "DrainArea",
    "Strahler",
    "Seg_ID",
    "Seg_order",
    "Max_DEM",
    "Min_DEM",
    "DA_Obs",
    "DA_error",
    "Obs_NM",
    "SRC_obs",
    "centroid_x",
    "centroid_y",
    "HRU_IsLake",
    "Landuse_ID",
    "Soil_ID",
    "Veg_ID",
    "HRU_Area",
    "HRU_ID",
    "LAND_USE_C",
    "VEG_C",
    "SOIL_PROF",
    "HRU_CenX",
    "HRU_CenY",
    "HRU_S_mean",
    "HRU_A_mean",
    "HRU_E_mean",
    "O_ID_1",
    "O_ID_2",
    "geometry",
]

LAND_USE_C_LAKE = "WATER"
LAND_USE_C_LAND = "Landuse_Land_HRU"
SOIL_PROF_LAKE = "LAKE"
SOIL_PROF_LAND = "Soil_Land_HRU"
VEG_C_LAKE = "WATER"
VEG_C_LAND = "Veg_Land_HRU"


@click.command()
@click.argument("input-file", type=click.Path(exists=True))
@click.option(
    "-o",
    "--output",
    type=click.Path(),
    help="Output shapefiles (will create a folder if no extension)",
)
def generate_hrus_from_routing_product(input_file, output):
    """
    Create a new HRU shapefile by splitting every subbasin row of a Routing Product V2.1 shapefile into at least a land HRU and possibly a lake HRU.

    Parameters
    ----------
    input_file : str
        Routing Product V2.1 shapefile (e.g. "drainage_region_0003_v2-1/finalcat_info_v2-1.shp").
    output : str
        Output shapefile (e.g. "hrus.shp").
    """

    def assign_hru_attributes(i_sub, i_hru, hru_id, is_lake_HRU):  # noqa: N803
        # fist copy subbasin attribute to hru table
        for sub_col in subbasin_info.columns:
            if sub_col == "geometry":
                subid = int(subbasin_info["SubId"].values[i_sub])
                # add a  new row in hruinfo
                hru_info.loc[i_hru, sub_col] = np.nan
                # copy geometry to hru
                hru_info.loc[[i_hru], sub_col] = subbasin_info.loc[[subid], sub_col].values
            else:
                hru_info.loc[i_hru, sub_col] = subbasin_info[sub_col].values[i_sub]

        hru_info.loc[i_hru, "HRU_ID"] = hru_id
        hru_info.loc[i_hru, "HRU_CenX"] = subbasin_info["centroid_x"].values[i_sub]
        hru_info.loc[i_hru, "HRU_CenY"] = subbasin_info["centroid_y"].values[i_sub]
        hru_info.loc[i_hru, "HRU_S_mean"] = subbasin_info["BasSlope"].values[i_sub]
        hru_info.loc[i_hru, "HRU_A_mean"] = subbasin_info["BasAspect"].values[i_sub]
        hru_info.loc[i_hru, "HRU_E_mean"] = subbasin_info["MeanElev"].values[i_sub]

        if is_lake_HRU:
            hru_info.loc[i_hru, "HRU_IsLake"] = 1
            hru_info.loc[i_hru, "Soil_ID"] = -1
            hru_info.loc[i_hru, "Veg_ID"] = -1
            hru_info.loc[i_hru, "O_ID_1"] = -1
            hru_info.loc[i_hru, "O_ID_2"] = -1
            hru_info.loc[i_hru, "Landuse_ID"] = -1
            hru_info.loc[i_hru, "HRU_Area"] = subbasin_info["LakeArea"].values[i_sub]
            hru_info.loc[i_hru, "LAND_USE_C"] = LAND_USE_C_LAKE
            hru_info.loc[i_hru, "SOIL_PROF"] = SOIL_PROF_LAKE
            hru_info.loc[i_hru, "VEG_C"] = VEG_C_LAKE
        else:
            hru_info.loc[i_hru, "HRU_IsLake"] = -1
            hru_info.loc[i_hru, "Soil_ID"] = 1
            hru_info.loc[i_hru, "Veg_ID"] = 1
            hru_info.loc[i_hru, "O_ID_1"] = 1
            hru_info.loc[i_hru, "O_ID_2"] = 1
            hru_info.loc[i_hru, "Landuse_ID"] = 1
            hru_info.loc[i_hru, "HRU_Area"] = subbasin_info["BasArea"].values[i_sub] - subbasin_info["LakeArea"].values[i_sub]
            hru_info.loc[i_hru, "LAND_USE_C"] = LAND_USE_C_LAND
            hru_info.loc[i_hru, "SOIL_PROF"] = SOIL_PROF_LAND
            hru_info.loc[i_hru, "VEG_C"] = VEG_C_LAND

    if Path(input_file).suffix == ".zip":
        input_file = f"zip://{input_file}"

    subbasin_info = gpd.read_file(input_file)

    if not output:
        p = Path(input_file)
        output_file = p.parent / f"{p.stem}_HRUs.shp"
    else:
        output_file = output

    # create an empty pandas table that each row will be 1 hru, columns will be
    # hru's attributes.
    hru_info = subbasin_info.copy(deep=True)

    # dissolve by subid, the subid will be set as index
    subbasin_info = subbasin_info.dissolve(by="SubId")
    # copy index back to new subid
    subbasin_info["SubId"] = subbasin_info.index

    max_subbsin_id = max(subbasin_info["SubId"].values)

    for col in COLUMN_NAMES_CONSTANT_HRU:
        if col != "geometry":
            hru_info[col] = np.nan

    i_hru = 0

    for i_sub in range(len(subbasin_info)):
        sub_id = subbasin_info["SubId"].values[i_sub]

        if subbasin_info["Lake_Cat"].values[i_sub] > 0:
            # hru id of land will be the corresponding subbasin id
            land_hru_id = sub_id
            assign_hru_attributes(
                i_sub=i_sub,
                i_hru=i_hru,
                hru_id=land_hru_id,
                is_lake_HRU=False,
            )

            i_hru += 1

            # hru id of lake will be subbasin id + max_subbsin_id + 10
            lake_hru_id = land_hru_id + max_subbsin_id + 10
            assign_hru_attributes(
                i_sub=i_sub,
                i_hru=i_hru,
                hru_id=lake_hru_id,
                is_lake_HRU=True,
            )
        else:
            assign_hru_attributes(
                i_sub=i_sub,
                i_hru=i_hru,
                hru_id=sub_id,
                is_lake_HRU=False,
            )

        i_hru += 1

    hru_info = hru_info.loc[hru_info["HRU_ID"] > 0]

    hru_info.to_file(output_file)

    click.echo(f"Created {output_file}")



import datetime as dt

import numpy as np
import pytest
import xarray as xr
from climpred import HindcastEnsemble

from ravenpy import Emulator, OutputReader
from ravenpy.utilities.forecasting import (
    climatology_esp,
    ensemble_prediction,
    hindcast_climatology_esp,
    warm_up,
)


def test_warm_up(minimal_emulator, tmp_path):
    conf = warm_up(minimal_emulator, duration=10, workdir=tmp_path)
    wup = OutputReader(path=tmp_path / "output")

    assert conf.start_date == minimal_emulator.start_date
    assert len(wup.hydrograph.q_sim.time) == 11
    assert conf.hru_state_variable_table != minimal_emulator.hru_state_variable_table


def test_climatology_esp(minimal_emulator, tmp_path):
    config = minimal_emulator.model_copy(deep=True)
    esp = climatology_esp(config, workdir=tmp_path, years=[1955, 1956])
    np.testing.assert_array_equal(esp.storage.member, [1955, 1956])
    assert len(esp.hydrograph.time) == minimal_emulator.duration + 1


def test_hindcast_climatology_esp(minimal_emulator, tmp_path, yangtze):
    config = minimal_emulator.model_copy(deep=True)
    hc = hindcast_climatology_esp(
        config,
        warm_up_duration=5,
        years=[1955, 1956, 1957],
        hindcast_years=[1990, 1991],
        workdir=tmp_path,
    )
    assert hc.sizes == {
        "lead": config.duration + 1,
        "init": 2,
        "member": 3,
        "nbasins": 1,
    }

    # Construct climpred HindcastEnsemble
    salmon_file = yangtze.fetch("raven-gr4j-cemaneige/Salmon-River-Near-Prince-George_meteo_daily.nc")
    qobs = xr.open_dataset(salmon_file).qobs

    qobs.name = "q_sim"
    hce = HindcastEnsemble(hc).add_observations(qobs)

    rank_histo_verif = hce.verify(
        metric="rank_histogram",
        comparison="m2o",
        dim=["member", "init"],
        alignment="same_inits",
    )
    assert rank_histo_verif.q_sim.shape[0] == config.duration + 1


def test_forecasting_GEPS(numeric_config, yangtze):
    """Test to perform a forecast using auto-queried ECCC data aggregated on THREDDS."""
    name, wup = numeric_config
    if name != "GR4JCN":
        pytest.skip("Test only for GR4JCN model.")

    geps = yangtze.fetch("eccc_forecasts/geps_watershed.nc")

    # Prepare a RAVEN model run using historical data, GR4JCN in this case.
    # This is a dummy run to get initial states. In a real forecast situation,
    # this run would end on the day before the forecast, but process is the same.
    wup.start_date = dt.datetime(2000, 1, 1)
    wup.duration = 30

    e = Emulator(wup)
    e.run()

    # Extract the final states that will be used as the next initial states
    conf = e.resume()

    # Set run parameters
    conf.start_date = "2020-12-10"
    conf.end_date = None
    conf.duration = 9

    # Collect test forecast data for location and climate model (20 members)
    nm = 20
    data_kwds = {
        "ALL": {
            "TimeShift": -0.25,
            "Latitude": conf.hrus[0].latitude,
            "Longitude": conf.hrus[0].longitude,
        },
        "PRECIP": {"Deaccumulate": True},
    }

    # Data types to extract from netCDF
    data_type = ["TEMP_AVE", "PRECIP"]

    # Run the 20 members
    out = ensemble_prediction(
        conf,
        geps,
        ens_dim="member",
        data_kwds=data_kwds,
        data_type=data_type,
    )

    # The model now has the forecast data generated and has 10 days of forecasts.
    assert len(out.hydrograph.time) == 10

    # Also see if GEPS has 20 members produced.
    assert len(out.hydrograph.q_sim.member) == nm

    # Check all members are different (checking snow because data in winter)
    assert len(set(out.storage.Snow.isel(time=-1).values)) == nm



from collections.abc import Sequence
from dataclasses import field, make_dataclass
from typing import Union

from pydantic import Field, field_validator, model_validator
from pydantic.dataclasses import dataclass
from pymbolic.primitives import Variable

import ravenpy.config.processes as p
from ravenpy.config import commands as rc
from ravenpy.config import options as o
from ravenpy.config.base import Sym, SymConfig
from ravenpy.config.commands import (
    HRU,
    PL,
    LandUseClasses,
    LandUseParameterList,
    SoilClasses,
    SoilParameterList,
    SoilProfiles,
    VegetationClasses,
)
from ravenpy.config.defaults import nc_attrs
from ravenpy.config.rvs import Config


P = dataclass(
    make_dataclass(
        "Params",
        [(f"X{i:02}", Sym, field(default=Variable(f"X{i:02}"))) for i in range(1, 35)],
    ),
    config=SymConfig,
)

# Bug: RavenC tries to write two variables with the same name `Snow Melt (Liquid) [mm]` to netCDF.


class OrganicHRU(HRU):
    hru_id: int = 1
    land_use_class: str = "FOREST"
    veg_class: str = "FOREST"
    soil_profile: str = "SOILP_ORG"
    aquifer_profile: str = "[NONE]"
    terrain_class: str = "[NONE]"


class BedRockHRU(HRU):
    hru_id: int = 2
    land_use_class: str = "FOREST"
    veg_class: str = "FOREST"
    soil_profile: str = "SOILP_BEDROCK"
    aquifer_profile: str = "[NONE]"
    terrain_class: str = "[NONE]"


class HRUs(rc.HRUs):
    """HRUs command for CanadianShield."""

    root: tuple[OrganicHRU, BedRockHRU]


class CanadianShield(Config):
    """
    The Canadian Shield model is a useful configuration of Raven for
    Canadian shield basins characterized by shallow soils atop rock,
    with ample exposed rock and lakes. It is a custom model for this
    type of region, but there is no reference to it in the literature.
    """

    params: P = P()
    hrus: HRUs = Field([OrganicHRU(), BedRockHRU()], alias="HRUs")
    netcdf_attribute: dict[str, str] = {"model_id": "CanadianShield"}
    sub_basins: rc.SubBasins = Field([rc.SubBasin()], alias="SubBasins")
    write_netcdf_format: bool = Field(True, alias="WriteNetcdfFormat")
    time_step: Union[float, str] = Field(1.0, alias="TimeStep")
    calendar: o.Calendar = Field("PROLEPTIC_GREGORIAN", alias="Calendar")
    land_use_classes: LandUseClasses = Field(
        [rc.LU(name="FOREST", impermeable_frac=0.0, forest_coverage=0.02345)],
        alias="LandUseClasses",
    )

    routing: o.Routing = Field("ROUTE_NONE", alias="Routing")
    catchment_route: o.CatchmentRoute = Field("ROUTE_TRI_CONVOLUTION", alias="CatchmentRouting")
    evaporation: o.Evaporation = Field(o.Evaporation.HARGREAVES_1985, alias="Evaporation")
    ow_evaporation: o.Evaporation = Field(o.Evaporation.HARGREAVES_1985, alias="OW_Evaporation")
    sw_canopy_correct: o.SWCanopyCorrect = Field(o.SWCanopyCorrect.STATIC, alias="SWCanopyCorrect")
    rain_snow_fraction: o.RainSnowFraction = Field(o.RainSnowFraction.DINGMAN, alias="RainSnowFraction")
    potential_melt_method: o.PotentialMeltMethod = Field(o.PotentialMeltMethod.DEGREE_DAY, alias="PotentialMeltMethod")
    precip_icept_frac: o.PrecipIceptFract = Field(o.PrecipIceptFract.LAI, alias="PrecipIceptFrac")
    soil_model: rc.SoilModel = Field(3, alias="SoilModel")
    monthly_interpolation_method: o.MonthlyInterpolationMethod = Field(o.MonthlyInterpolationMethod.LINEAR_MID, alias="MonthlyInterpolationMethod")
    hydrologic_processes: Sequence[Union[rc.Process, p.Conditional]] = Field(
        [
            p.SnowRefreeze(algo="FREEZE_DEGREE_DAY", source="SNOW_LIQ", to="SNOW"),
            p.Precipitation(algo="PRECIP_RAVEN", source="ATMOS_PRECIP", to="MULTIPLE"),
            p.CanopyEvaporation(algo="CANEVP_MAXIMUM", source="CANOPY", to="ATMOSPHERE"),
            p.CanopySublimation(algo="CANEVP_MAXIMUM", source="CANOPY_SNOW", to="ATMOSPHERE"),
            p.SnowBalance(algo="SNOBAL_TWO_LAYER", source="MULTIPLE", to="MULTIPLE"),
            p.Abstraction(algo="ABST_FILL", source="PONDED_WATER", to="DEPRESSION"),
            p.OpenWaterEvaporation(algo="OPEN_WATER_EVAP", source="DEPRESSION", to="ATMOSPHERE"),
            p.Infiltration(algo="INF_HBV", source="PONDED_WATER", to="MULTIPLE"),
            p.Interflow(algo="INTERFLOW_PRMS", source="SOIL[0]", to="SURFACE_WATER"),
            p.Baseflow(algo="BASE_POWER_LAW", source="SOIL[1]", to="SURFACE_WATER"),
            p.Conditional(kind="HRU_TYPE", op="IS", value="SOIL_ORG"),
            p.Baseflow(algo="BASE_POWER_LAW", source="SOIL[2]", to="SURFACE_WATER"),
            p.Percolation(algo="PERC_GAWSER", source="SOIL[0]", to="SOIL[1]"),
            p.Conditional(kind="HRU_TYPE", op="IS", value="SOIL_ORG"),
            p.Percolation(algo="PERC_GAWSER", source="SOIL[1]", to="SOIL[2]"),
            p.Conditional(kind="HRU_TYPE", op="IS", value="SOIL_ORG"),
            p.Percolation(algo="PERC_GAWSER", source="SOIL[0]", to="SOIL[2]"),
            p.Conditional(kind="HRU_TYPE", op="IS", value="SOIL_BEDROCK"),
            p.SoilEvaporation(algo="SOILEVAP_ROOT", source="SOIL[0]", to="ATMOSPHERE"),
        ],
        alias="HydrologicProcesses",
    )

    soil_classes: SoilClasses = Field(
        [{"name": "TOPSOIL"}, {"name": "VADOSE"}, {"name": "FRACBEDROCK"}],
        alias="SoilClasses",
    )
    soil_profiles: SoilProfiles = Field(
        [
            {"name": "LAKE"},
            {"name": "ROCK"},
            {
                "name": "SOILP_ORG",
                "soil_classes": ("TOPSOIL", "VADOSE", "FRACBEDROCK"),
                "thicknesses": (P.X01, P.X02, P.X03),
            },
            {
                "name": "SOILP_BEDROCK",
                "soil_classes": ("TOPSOIL", "VADOSE", "FRACBEDROCK"),
                "thicknesses": (P.X01, 0, P.X03),
            },
        ],
        alias="SoilProfiles",
    )
    vegetation_classes: VegetationClasses = Field(
        [{"name": "FOREST", "max_ht": 5, "max_lai": 5, "max_leaf_cond": 5}],
        alias="VegetationClasses",
    )

    global_parameter: dict = Field(
        {
            "SNOW_SWI": P.X15,
            "SNOW_SWI_MIN": P.X16,
            "SNOW_SWI_MAX": P.X17,
            "SWI_REDUCT_COEFF": P.X18,
            "RAINSNOW_TEMP": P.X19,
            "RAINSNOW_DELTA": P.X20,
            "MAX_SWE_SURFACE": P.X21,
            "TOC_MULTIPLIER": P.X22,
        }
    )
    soil_parameter_list: SoilParameterList = Field(
        {
            "parameters": (
                "POROSITY",
                "HBV_BETA",
                "BASEFLOW_COEFF",
                "BASEFLOW_N",
                "MAX_INTERFLOW_RATE",
                "FIELD_CAPACITY",
                "SAT_WILT",
                "MAX_PERC_RATE",
                "PET_CORRECTION",
            ),
            "pl": [
                PL(
                    name="TOPSOIL",
                    values=(1.0, P.X07, 0.0, 0.0, P.X12, P.X06, P.X05, P.X13, P.X04),
                ),
                PL(name="VADOSE", values=(1.0, 0.0, P.X08, P.X10, 0, 0, 0, P.X14, 0)),
                PL(name="FRACBEDROCK", values=(1.0, 0.0, P.X09, P.X11, 0, 0, 0, 0, 0)),
            ],
        },
        alias="SoilParameterList",
    )
    land_use_parameter_list: LandUseParameterList = Field(
        {
            "parameters": (
                "FOREST_SPARSENESS",
                "MELT_FACTOR",
                "DD_MELT_TEMP",
                "REFREEZE_FACTOR",
                "DEP_MAX",
                "OW_PET_CORR",
            ),
            "pl": [PL(name="FOREST", values=(0, P.X25, P.X24, P.X23, P.X26, P.X27))],
        },
        alias="LandUseParameterList",
    )
    vegetation_parameter_list: rc.VegetationParameterList = Field(
        {
            "parameters": (
                "SVF_EXTINCTION",
                "SAI_HT_RATIO",
                "RAIN_ICEPT_FACT",
                "SNOW_ICEPT_FACT",
                "MAX_CAPACITY",
                "MAX_SNOW_CAPACITY",
            ),
            "pl": [PL(name="FOREST", values=(0.5, 1, P.X28, P.X29, P.X30, P.X31))],
        },
        alias="VegetationParameterList",
    )
    seasonal_relative_lai: rc.SeasonalRelativeLAI = Field(rc.SeasonalRelativeLAI(), alias="SeasonalRelativeLAI")
    seasonal_relative_height: rc.SeasonalRelativeHeight = Field(rc.SeasonalRelativeHeight(), alias="SeasonalRelativeHeight")

    hru_state_variable_table: rc.HRUStateVariableTable = Field(
        [
            rc.HRUState(
                hru_id=1,
                data={
                    "SOIL[0]": P.X01 * 500,
                    "SOIL[1]": P.X02 * 500,
                    "SOIL[2]": P.X03 * 500,
                },
            ),
            rc.HRUState(
                hru_id=2,
                data={
                    "SOIL[0]": P.X01 * 500,
                    "SOIL[1]": P.X02 * 500,
                    "SOIL[2]": P.X03 * 500,
                },
            ),
        ],
        alias="HRUStateVariableTable",
    )

    @field_validator("hrus", mode="before")
    @classmethod
    def equal_area(cls, hrus):
        if len(hrus) != 2:
            raise ValueError(f"CanadianShield must have two HRUs, one organic and one bedrock, has {len(hrus)}.")

        areas = []
        for hru in hrus:
            if isinstance(hru, dict):
                areas.append(hru["area"])
            else:
                areas.append(hru.area)

        if len(set(areas)) > 1:
            raise ValueError(f"HRUs must have equal areas: {areas}")

        return hrus

    _nc_attrs = field_validator("netcdf_attribute")(nc_attrs)
    _initialized: bool = False

    @model_validator(mode="after")
    def _init(self):
        if not self._initialized:
            hrus = self.hrus.root
            hrus[0].area *= self.params.X34
            hrus[1].area *= 1 - self.params.X34

            if self.gauge:
                for gauge in self.gauge:
                    gauge.rain_correction = self.params.X32
                    gauge.snow_correction = self.params.X33

            self._initialized = True

        return self



"""
GeoServer interaction operations.

Working assumptions for this module:
* Point coordinates are passed as shapely.geometry.Point instances.
* BBox coordinates are passed as (lon1, lat1, lon2, lat2).
* Shapes (polygons) are passed as shapely.geometry.shape parsable objects.
* All functions that require a CRS have a CRS argument with a default set to WGS84.
* GEOSERVER_URL points to the GeoServer instance hosting all files.
* For legacy reasons, we also accept the `GEO_URL` environment variable.

TODO: Refactor to remove functions that are just 2-lines of code.
For example, many function's logic essentially consists in creating the layer name.
We could have a function that returns the layer name, and then other functions expect the layer name.
"""

import json
import os
import warnings
from collections.abc import Iterable, Sequence
from pathlib import Path
from typing import Optional, Union
from urllib.parse import urlencode, urljoin

import urllib3

from . import gis_import_error_message


try:
    import fiona
    import geopandas as gpd
    import pandas as pd
    from lxml import etree
    from owslib.fes import PropertyIsLike
    from owslib.fes2 import Intersects
    from owslib.gml import Point
    from owslib.wcs import WebCoverageService
    from owslib.wfs import WebFeatureService
except (ImportError, ModuleNotFoundError) as err:
    msg = gis_import_error_message.format(Path(__file__).stem)
    raise ImportError(msg) from err

from .geo import determine_upstream_ids


# Can be set at runtime with `$ env RAVENPY_GEOSERVER_URL=https://xx.yy.zz/geoserver/ ...`.
# For legacy reasons, we also accept the `GEO_URL` environment variable.
HOST_URL = os.getenv("RAVENPY_HOST_URL", os.getenv("HOST_URL", "https://pavics.ouranos.ca/"))
GEOSERVER_URL = os.getenv(
    "RAVENPY_GEOSERVER_URL",
    os.getenv("GEO_URL", f"{HOST_URL}/geoserver/"),
)
if not GEOSERVER_URL.endswith("/"):
    GEOSERVER_URL = f"{GEOSERVER_URL}/"

# We store the contour of different HydroBASINS domains
hybas_dir = Path(__file__).parent.parent / "data" / "hydrobasins_domains"
hybas_pat = "hybas_lake_{domain}_lev01_v1c.zip"

# This could be inferred from existing files in hybas_dir
hybas_regions = ["na", "ar"]
hybas_domains = {dom: hybas_dir / hybas_pat.format(domain=dom) for dom in hybas_regions}


def _fix_server_url(server_url: str) -> str:
    if not server_url.endswith("/"):
        warnings.warn("The url should end with a slash. Appending it to the url.", stacklevel=2)
        return f"{server_url}/"
    return server_url


def _get_location_wfs(
    bbox: Optional[
        tuple[
            Union[str, float, int],
            Union[str, float, int],
            Union[str, float, int],
            Union[str, float, int],
        ]
    ] = None,
    point: Optional[
        tuple[
            Union[str, float, int],
            Union[str, float, int],
        ]
    ] = None,
    *,
    layer: str,
    geoserver: str = GEOSERVER_URL,
) -> dict:
    """
    Return leveled features from a hosted data set using bounding box coordinates and WFS 1.1.0 protocol.

    For geographic rasters, subsetting is based on WGS84 (Long, Lat) boundaries. If not geographic, subsetting based
    on projected coordinate system (Easting, Northing) boundaries.

    Parameters
    ----------
    bbox : Optional[Tuple[Union[str, float, int], Union[str, float, int], Union[str, float, int], Union[str, float, int]]]
        Geographic coordinates of the bounding box (left, down, right, up).
    point : Optional[Tuple[Union[str, float, int], Union[str, float, int]]]
        Geographic coordinates of an intersecting point (lon, lat).
    layer : str
        The WFS/WMS layer name requested.
    geoserver : str
        The address of the geoserver housing the layer to be queried. Default: https://pavics.ouranos.ca/geoserver/.

    Returns
    -------
    dict
        A GeoJSON-derived dictionary of vector features (FeatureCollection).
    """
    geoserver = _fix_server_url(geoserver)

    wfs = WebFeatureService(url=urljoin(geoserver, "wfs"), version="2.0.0", timeout=30)

    if bbox and point:
        raise NotImplementedError("Provide either 'bbox' or 'point'.")
    if bbox:
        kwargs = dict(bbox=bbox)
    elif point:
        p = Point(
            id="feature",
            srsName="http://www.opengis.net/gml/srs/epsg.xml#4326",
            pos=point,
        )
        f = Intersects(propertyname="the_geom", geometry=p)
        intersects = f.toXML()
        kwargs = dict(filter=intersects)
    else:
        raise ValueError("Either 'bbox' or 'point' must be provided.")

    resp = wfs.getfeature(typename=layer, outputFormat="application/json", method="POST", **kwargs)

    data = json.loads(resp.read())
    return data


def _get_feature_attributes_wfs(
    attribute: Sequence[str],
    layer: str,
    geoserver: str = GEOSERVER_URL,
) -> str:
    """
    Return WFS GetFeature URL request for attribute values.

    Making this request will return a JSON response.

    Parameters
    ----------
    attribute : list
        Attribute/field names.
    layer : str
        Name of geographic layer queried.
    geoserver : str
        The address of the geoserver housing the layer to be queried.
        Default: https://pavics.ouranos.ca/geoserver/.

    Returns
    -------
    str
        WFS request URL.

    Notes
    -----
    Non-existent attributes will raise a cryptic DriverError from fiona.
    """
    host = _fix_server_url(HOST_URL)
    geoserver = _fix_server_url(geoserver)

    params = dict(
        service="WFS",
        version="2.0.0",
        request="GetFeature",
        typename=layer,
        outputFormat="application/json",
        propertyName=",".join(attribute),
    )
    url = urljoin(geoserver, "wfs") + "?" + urlencode(params)
    http = urllib3.PoolManager()
    response = http.request("GET", url)

    return urljoin(host, response.url)


def _filter_feature_attributes_wfs(
    attribute: str,
    value: Union[str, float, int],
    layer: str,
    geoserver: str = GEOSERVER_URL,
) -> str:
    """
    Return WFS GetFeature URL request filtering geographic features based on a property's value.

    Parameters
    ----------
    attribute : str
        Attribute/field name.
    value: Union[str, float, int]
        Value for attribute queried.
    layer : str
        Name of geographic layer queried.
    geoserver : str
        The address of the geoserver housing the layer to be queried.
        Default: https://pavics.ouranos.ca/geoserver/.

    Returns
    -------
    str
      WFS request URL.
    """
    host = _fix_server_url(HOST_URL)
    geoserver = _fix_server_url(geoserver)

    try:
        attribute = str(attribute)
        value = str(value)

    except ValueError as e:
        raise ValueError("Unable to cast attribute/filter to string.") from e

    filter_request = PropertyIsLike(propertyname=attribute, literal=value, wildCard="*")
    filterxml = etree.tostring(filter_request.toXML()).decode("utf-8")
    params = dict(
        service="WFS",
        version="1.1.0",
        request="GetFeature",
        typename=layer,
        outputFormat="application/json",
        filter=filterxml,
    )

    url = urljoin(geoserver, "wfs") + "?" + urlencode(params)
    http = urllib3.PoolManager()
    response = http.request("GET", url)

    return urljoin(host, response.url)


def get_raster_wcs(
    coordinates: Union[Iterable, Sequence[Union[float, str]]],
    geographic: bool = True,
    *,
    layer: str,
    geoserver: str = GEOSERVER_URL,
) -> bytes:
    """
    Return a subset of a raster image from the local GeoServer via WCS 2.0.1 protocol.

    For geographic raster grids, subsetting is based on WGS84 (Long, Lat) boundaries.
    If not geographic, subsetting based on projected coordinate system (Easting, Northing) boundaries.

    Parameters
    ----------
    coordinates : Sequence of int or float or str
        Geographic coordinates of the bounding box (left, down, right, up).
    geographic : bool
        If True, uses "Long" and "Lat" in WCS call. Otherwise, uses "E" and "N".
    layer : str
        Layer name of raster exposed on GeoServer instance, e.g. 'public:CEC_NALCMS_LandUse_2010'.
    geoserver : str
        The address of the geoserver housing the layer to be queried.
        Default: https://pavics.ouranos.ca/geoserver/.

    Returns
    -------
    bytes
        A GeoTIFF array.
    """
    geoserver = _fix_server_url(geoserver)

    (left, down, right, up) = coordinates

    if geographic:
        x, y = "Long", "Lat"
    else:
        x, y = "E", "N"

    wcs = WebCoverageService(url=urljoin(geoserver, "ows"), version="2.0.1")

    try:
        resp = wcs.getCoverage(
            identifier=[layer],
            format="image/tiff",
            subsets=[(x, left, right), (y, down, up)],
            timeout=120,
        )

    except Exception:
        raise

    data = resp.read()

    try:
        etree.fromstring(data)  # noqa: S320
        # The response is an XML file describing the server error.
        raise ChildProcessError(data)

    except etree.XMLSyntaxError:
        # The response is the DEM array.
        return data


# ~~~~ HydroBASINS functions ~~~~ #


def hydrobasins_upstream(feature: dict, domain: str) -> pd.DataFrame:
    """
    Return a list of HydroBASINS features located upstream.

    Parameters
    ----------
    feature : dict
        Basin feature attributes, including the fields ["HYBAS_ID", "NEXT_DOWN", "MAIN_BAS"].
    domain : {"na", "ar"}
        Domain of the feature, North America or Arctic.

    Returns
    -------
    pd.Series
        Basins ids including `fid` and its upstream contributors.
    """
    basin_field = "HYBAS_ID"
    downstream_field = "NEXT_DOWN"
    basin_family = "MAIN_BAS"

    # This does not work with `wfs.getfeature`. No filtering occurs when asking for specific attributes.
    # wfs = WebFeatureService(url=urljoin(geoserver, "wfs"), version="2.0.0", timeout=30)
    # layer = f"public:USGS_HydroBASINS_{'lake_' if lakes else ''}{domain}_lev{str(level).zfill(2)}"
    # filter = PropertyIsEqualTo(propertyname=basin_family, literal=feature[basin_family])

    # Fetch all features in the same basin
    request_url = filter_hydrobasins_attributes_wfs(attribute=basin_family, value=feature[basin_family], domain=domain)
    df = gpd.read_file(filename=request_url, engine="pyogrio")

    # Filter upstream watersheds
    return determine_upstream_ids(
        fid=feature[basin_field],
        df=df,
        basin_field=basin_field,
        downstream_field=downstream_field,
    )


def hydrobasins_aggregate(gdf: pd.DataFrame) -> pd.DataFrame:
    """
    Aggregate multiple HydroBASINS watersheds into a single geometry.

    Parameters
    ----------
    gdf : pd.DataFrame
        Watershed attributes indexed by HYBAS_ID.

    Returns
    -------
    pd.DataFrame
    """
    i0 = gdf.index[0]

    # TODO: Review. Not sure it all makes sense. --> Looks fine to me? (TJS)
    def aggfunc(x):
        if x.name in ["COAST", "DIST_MAIN", "DIST_SINK"]:
            return x.min()
        elif x.name in ["SUB_AREA", "LAKE"]:
            return x.sum()
        else:
            return x.loc[i0]

    # Buffer function to fix invalid geometries
    gdf["geometry"] = gdf.buffer(0)

    return gdf.dissolve(by="MAIN_BAS", aggfunc=aggfunc)


def select_hybas_domain(
    bbox: Optional[tuple[Union[int, float], Union[int, float], Union[int, float], Union[int, float]]] = None,
    point: Optional[tuple[Union[int, float], Union[int, float]]] = None,
) -> str:
    """
    Provided a given coordinate or boundary box, return the domain name of the geographic region the coordinate is located within.

    Parameters
    ----------
    bbox : Optional[Tuple[Union[float, int], Union[float, int], Union[float, int], Union[float, int]]]
        Geographic coordinates of the bounding box (left, down, right, up).
    point : Optional[Tuple[Union[float, int], Union[float, int]]]
        Geographic coordinates of an intersecting point (lon, lat).

    Returns
    -------
    str
        The domain that the coordinate falls within. Possible results: "na", "ar".
    """
    if bbox and point:
        raise NotImplementedError("Provide either 'bbox' or 'point'.")
    if point:
        bbox = point * 2

    for dom, fn in hybas_domains.items():
        with Path(fn).open("rb") as f:
            zf = fiona.io.ZipMemoryFile(f)
            coll = zf.open(fn.stem + ".shp")
            for feat in coll.filter(bbox=bbox):
                if isinstance(feat, fiona.Feature):
                    return dom

    raise LookupError(f"Could not find feature containing bbox: {bbox}.")


def filter_hydrobasins_attributes_wfs(
    attribute: str,
    value: Union[str, float, int],
    domain: str,
    geoserver: str = GEOSERVER_URL,
) -> str:
    """
    Return a URL that formats and returns a remote GetFeatures request from the USGS HydroBASINS dataset.

    For geographic raster grids, subsetting is based on WGS84 (Long, Lat) boundaries.
    If not geographic, subsetting based on projected coordinate system (Easting, Northing) boundaries.

    Parameters
    ----------
    attribute : str
        Attribute/field to be queried.
    value : str or float or int
        Value for attribute queried.
    domain : {"na", "ar"}
        The domain of the HydroBASINS data.
    geoserver : str
        The address of the geoserver housing the layer to be queried. Default: https://pavics.ouranos.ca/geoserver/.

    Returns
    -------
    str
        URL to the GeoJSON-encoded WFS response.
    """
    geoserver = _fix_server_url(geoserver)

    lakes = True
    level = 12

    layer = f"public:USGS_HydroBASINS_{'lake_' if lakes else ''}{domain}_lev{str(level).zfill(2)}"
    q = _filter_feature_attributes_wfs(attribute=attribute, value=value, layer=layer, geoserver=geoserver)

    return q


def get_hydrobasins_location_wfs(
    coordinates: tuple[
        Union[str, float, int],
        Union[str, float, int],
    ],
    domain: str,
    geoserver: str = GEOSERVER_URL,
) -> dict[str, Union[str, int, float]]:
    """
    Return features from the USGS HydroBASINS data set using bounding box coordinates.

    For geographic raster grids, subsetting is based on WGS84 (Long, Lat) boundaries.
    If not geographic, subsetting based on projected coordinate system (Easting, Northing) boundaries.

    Parameters
    ----------
    coordinates : Tuple[str or float or int, str or float or int]
        Geographic coordinates of the bounding box (left, down, right, up).
    domain : {"na", "ar"}
        The domain of the HydroBASINS data.
    geoserver : str
        The address of the geoserver housing the layer to be queried. Default: https://pavics.ouranos.ca/geoserver/.

    Returns
    -------
    dict
        A GeoJSON-encoded vector feature.
    """
    geoserver = _fix_server_url(geoserver)

    lakes = True
    level = 12
    layer = f"public:USGS_HydroBASINS_{'lake_' if lakes else ''}{domain}_lev{str(level).zfill(2)}"
    if not Point and not Intersects:
        data = _get_location_wfs(bbox=coordinates * 2, layer=layer, geoserver=geoserver)
    else:
        data = _get_location_wfs(point=coordinates, layer=layer, geoserver=geoserver)

    return data


# ~~~~ Hydro Routing ~~~~ #


def hydro_routing_upstream(
    fid: Union[str, float, int],
    level: int = 12,
    lakes: str = "1km",
    geoserver: str = GEOSERVER_URL,
) -> pd.Series:
    """
    Return a list of hydro routing features located upstream.

    Parameters
    ----------
    fid : str or float or int
        Basin feature ID code of the downstream feature.
    level : int
        Level of granularity requested for the lakes vector (range(7,13)). Default: 12.
    lakes : {"1km", "all"}
        Query the version of dataset with lakes under 1km in width removed ("1km") or return all lakes ("all").
    geoserver : str
        The address of the geoserver housing the layer to be queried.
        Default: https://pavics.ouranos.ca/geoserver/.

    Returns
    -------
    pd.Series
        Basins ids including `fid` and its upstream contributors.
    """
    geoserver = _fix_server_url(geoserver)

    wfs = WebFeatureService(url=urljoin(geoserver, "wfs"), version="2.0.0", timeout=30)
    layer = f"public:routing_{lakes}Lakes_{str(level).zfill(2)}"

    # Get attributes necessary to identify upstream watersheds
    resp = wfs.getfeature(
        typename=layer,
        propertyname=["SubId", "DowSubId"],
        outputFormat="application/json",
    )
    df = gpd.read_file(resp)

    # Identify upstream features
    df_upstream = determine_upstream_ids(
        fid=fid,
        df=df,
        basin_field="SubId",
        downstream_field="DowSubId",
    )

    # Fetch upstream features
    resp = wfs.getfeature(
        typename=layer,
        featureid=df_upstream["id"].tolist(),
        outputFormat="application/json",
    )

    return gpd.read_file(resp.read().decode())


def get_hydro_routing_attributes_wfs(
    attribute: Sequence[str],
    level: int = 12,
    lakes: str = "1km",
    geoserver: str = GEOSERVER_URL,
) -> str:
    """
    Return a URL that formats and returns a remote GetFeatures request from hydro routing dataset.

    For geographic rasters, subsetting is based on WGS84 (Long, Lat) boundaries.
    If not geographic, subsetting based on projected coordinate system (Easting, Northing) boundaries.

    Parameters
    ----------
    attribute : list
        Attributes/fields to be queried.
    level : int
        Level of granularity requested for the lakes vector (range(7,13)). Default: 12.
    lakes : {"1km", "all"}
        Query the version of dataset with lakes under 1km in width removed ("1km") or return all lakes ("all").
    geoserver : str
        The address of the geoserver housing the layer to be queried.
        Default: https://pavics.ouranos.ca/geoserver/.

    Returns
    -------
    str
        URL to the GeoJSON-encoded WFS response.
    """
    geoserver = _fix_server_url(geoserver)

    layer = f"public:routing_{lakes}Lakes_{str(level).zfill(2)}"
    return _get_feature_attributes_wfs(attribute=attribute, layer=layer, geoserver=geoserver)


def filter_hydro_routing_attributes_wfs(
    attribute: str,
    value: Union[str, float, int],
    level: int = 12,
    lakes: str = "1km",
    geoserver: str = GEOSERVER_URL,
) -> str:
    """
    Return a URL that formats and returns a remote GetFeatures request from hydro routing dataset.

    For geographic rasters, subsetting is based on WGS84 (Long, Lat) boundaries. If not geographic, subsetting based
    on projected coordinate system (Easting, Northing) boundaries.

    Parameters
    ----------
    attribute : list
        Attributes/fields to be queried.
    value : str or int or float
        The requested value for the attribute.
    level : int
        Level of granularity requested for the lakes vector (range(7,13)). Default: 12.
    lakes : {"1km", "all"}
        Query the version of dataset with lakes under 1km in width removed ("1km") or return all lakes ("all").
    geoserver : str
        The address of the geoserver housing the layer to be queried.
        Default: https://pavics.ouranos.ca/geoserver/.

    Returns
    -------
    str
        URL to the GeoJSON-encoded WFS response.
    """
    geoserver = _fix_server_url(geoserver)

    layer = f"public:routing_{lakes}Lakes_{str(level).zfill(2)}"
    return _filter_feature_attributes_wfs(attribute=attribute, value=value, layer=layer, geoserver=geoserver)


def get_hydro_routing_location_wfs(
    coordinates: tuple[
        Union[int, float, str],
        Union[str, float, int],
    ],
    lakes: str,
    level: int = 12,
    geoserver: str = GEOSERVER_URL,
) -> dict:
    """
    Return features from the hydro routing data set using bounding box coordinates.

    For geographic rasters, subsetting is based on WGS84 (Long, Lat) boundaries. If not geographic, subsetting based
    on projected coordinate system (Easting, Northing) boundaries.

    Parameters
    ----------
    coordinates : Tuple[str or float or int, str or float or int]
        Geographic coordinates of the bounding box (left, down, right, up).
    lakes : {"1km", "all"}
        Query the version of dataset with lakes under 1km in width removed ("1km") or return all lakes ("all").
    level : int
        Level of granularity requested for the lakes vector (range(7,13)). Default: 12.
    geoserver : str
        The address of the geoserver housing the layer to be queried.
        Default: https://pavics.ouranos.ca/geoserver/.

    Returns
    -------
    dict
        A GeoJSON-derived dictionary of vector features (FeatureCollection).
    """
    geoserver = _fix_server_url(geoserver)

    layer = f"public:routing_{lakes}Lakes_{str(level).zfill(2)}"

    if not Point and not Intersects:
        data = _get_location_wfs(bbox=coordinates * 2, layer=layer, geoserver=geoserver)
    else:
        data = _get_location_wfs(point=coordinates, layer=layer, geoserver=geoserver)

    return data



"""Tools for searching for and acquiring test data."""

from __future__ import annotations
import importlib.metadata as ilm
import importlib.resources as ilr
import logging
import os
import re
import tempfile
import time
import warnings
from collections.abc import Callable
from datetime import datetime as dt
from functools import wraps
from io import StringIO
from pathlib import Path
from shutil import copytree
from typing import IO, Any, TextIO
from urllib.error import HTTPError, URLError
from urllib.parse import urljoin, urlparse
from urllib.request import urlretrieve

from filelock import FileLock
from packaging.version import Version
from xarray import Dataset
from xarray import open_dataset as _open_dataset
from xclim.testing.utils import show_versions as _show_versions

import ravenpy


try:
    import pooch
except ImportError:
    warnings.warn(
        "The `pooch` library is not installed. The default cache directory for testing data will not be set.",
        stacklevel=2,
    )
    pooch = None

LOGGER = logging.getLogger("ravenpy.testing.utils")

__all__ = [
    "TESTDATA_BRANCH",
    "TESTDATA_CACHE_DIR",
    "TESTDATA_REPO_URL",
    "audit_url",
    "default_testdata_cache",
    "default_testdata_repo_url",
    "default_testdata_version",
    "gather_testing_data",
    "get_file",
    "open_dataset",
    "populate_testing_data",
    "show_versions",
    "testing_setup_warnings",
    "yangtze",
]

default_testdata_version = "v2025.6.12"
"""Default version of the testing data to use when fetching datasets."""

default_testdata_repo_url = "https://raw.githubusercontent.com/Ouranosinc/raven-testdata/"
"""Default URL of the testing data repository to use when fetching datasets."""

try:
    default_testdata_cache = Path(pooch.os_cache("raven-testdata"))
    """Default location for the testing data cache."""
except (AttributeError, TypeError):
    default_testdata_cache = None

TESTDATA_REPO_URL = str(os.getenv("RAVEN_TESTDATA_REPO_URL", default_testdata_repo_url))
"""
Sets the URL of the testing data repository to use when fetching datasets.

Notes
-----
When running tests locally, this can be set for both `pytest` and `tox` by exporting the variable:

.. code-block:: console

    $ export RAVEN_TESTDATA_REPO_URL="https://github.com/my_username/raven-testdata"

or setting the variable at runtime:

.. code-block:: console

    $ env RAVEN_TESTDATA_REPO_URL="https://github.com/my_username/raven-testdata" pytest
"""

TESTDATA_BRANCH = str(os.getenv("RAVEN_TESTDATA_BRANCH", default_testdata_version))
"""
Sets the branch of the testing data repository to use when fetching datasets.

Notes
-----
When running tests locally, this can be set for both `pytest` and `tox` by exporting the variable:

.. code-block:: console

    $ export RAVEN_TESTDATA_BRANCH="my_testing_branch"

or setting the variable at runtime:

.. code-block:: console

    $ env RAVEN_TESTDATA_BRANCH="my_testing_branch" pytest
"""

TESTDATA_CACHE_DIR = os.getenv("RAVEN_TESTDATA_CACHE_DIR", default_testdata_cache)
"""
Sets the directory to store the testing datasets.

If not set, the default location will be used (based on ``platformdirs``, see :func:`pooch.os_cache`).

Notes
-----
When running tests locally, this can be set for both `pytest` and `tox` by exporting the variable:

.. code-block:: console

    $ export RAVEN_TESTDATA_CACHE_DIR="/path/to/my/data"

or setting the variable at runtime:

.. code-block:: console

    $ env RAVEN_TESTDATA_CACHE_DIR="/path/to/my/data" pytest
"""


def show_versions(
    file: os.PathLike | StringIO | TextIO | None = None,
    deps: list | None = None,
) -> str | None:
    """
    Print the versions of RavenPy and its dependencies.

    Parameters
    ----------
    file : {os.PathLike, StringIO, TextIO}, optional
        If provided, prints to the given file-like object. Otherwise, returns a string.
    deps : list, optional
        A list of dependencies to gather and print version information from. Otherwise, print `RavenPy` dependencies.

    Returns
    -------
    str or None
        The formatted version information if `file` is not provided, otherwise None.
    """

    def _get_ravenpy_dependencies():
        ravenpy_metadata = ilm.metadata("ravenpy")
        requires = ravenpy_metadata.get_all("Requires-Dist")
        requires = [re.match(r"^[A-Za-z0-9_.\-]+", req).group(0) for req in requires if re.match(r"^[A-Za-z0-9_.\-]+", req)]
        sorted_deps = sorted(list(set(requires) - {"ravenpy"}))

        return ["ravenpy"] + sorted_deps

    if deps is None:
        deps = _get_ravenpy_dependencies()

    return _show_versions(file=file, deps=deps)


# Test Data Utilities ###


def testing_setup_warnings():
    """Warn users about potential incompatibilities between RavenPy and raven-testdata versions."""
    if re.match(r"^\d+\.\d+\.\d+$", ravenpy.__version__) and TESTDATA_BRANCH != default_testdata_version:
        # This does not need to be emitted on GitHub Workflows and ReadTheDocs
        if not os.getenv("CI") and not os.getenv("READTHEDOCS"):
            warnings.warn(
                f"`RavenPy` stable ({ravenpy.__version__}) is running tests against a non-default "
                f"branch of the testing data. It is possible that changes to the testing data may "
                f"be incompatible with some assertions in this version. "
                f"Please be sure to check {TESTDATA_REPO_URL} for more information.",
                stacklevel=2,
            )

    if re.match(r"^v\d+\.\d+\.\d+", TESTDATA_BRANCH):
        # Find the date of the last modification of RavenPy source files to generate a calendar version
        install_date = dt.strptime(
            time.ctime(Path(ravenpy.__file__).stat().st_mtime),
            "%a %b %d %H:%M:%S %Y",
        )
        install_calendar_version = f"{install_date.year}.{install_date.month}.{install_date.day}"

        if Version(TESTDATA_BRANCH) > Version(install_calendar_version):
            warnings.warn(
                f"The installation date of `RavenPy` ({install_date.ctime()}) "
                f"predates the last release of testing data ({TESTDATA_BRANCH}). "
                "It is very likely that the testing data is incompatible with this build of `RavenPy`.",
                stacklevel=2,
            )


def load_registry(
    branch: str = TESTDATA_BRANCH,
    repo: str = TESTDATA_REPO_URL,
    force_download: bool = False,
) -> dict[str, str]:
    """
    Load the registry file for the test data.

    Parameters
    ----------
    branch : str
        Branch of the repository to use when fetching testing datasets.
    repo : str
        URL of the repository to use when fetching testing datasets.
    force_download : bool
        If True, force the download of the registry file even if it already exists.

    Returns
    -------
    dict
        Dictionary of filenames and hashes.
    """

    def load_registry_from_file(
        _registry_file: str | Path,
    ) -> dict[str, str]:
        """Load the registry from a file."""
        with Path(_registry_file).open(encoding="utf-8") as f:
            return {line.split()[0]: line.split()[1] for line in f}

    if not repo.endswith("/"):
        repo = f"{repo}/"
    remote_registry = audit_url(
        urljoin(
            urljoin(repo, branch if branch.endswith("/") else f"{branch}/"),
            "data/registry.txt",
        )
    )

    if repo != default_testdata_repo_url:
        external_repo_name = urlparse(repo).path.split("/")[-2]
        external_branch_name = branch.split("/")[-1]
        testing_folder = Path(str(ilr.files("ravenpy").joinpath("testing")))
        registry_file = testing_folder.joinpath(f"registry.{external_repo_name}.{external_branch_name}.txt")
        lockfile = testing_folder.joinpath(".lock")
        with FileLock(lockfile):
            if not registry_file.exists():
                urlretrieve(remote_registry, registry_file)  # noqa: S310
        lockfile.unlink(missing_ok=True)

    elif branch != default_testdata_version:
        # If force_download is True, download to a transient directory for testing purposes
        if force_download:
            with tempfile.TemporaryDirectory() as tmp_dir:
                custom_registry_folder = Path(tmp_dir).joinpath("testing", branch)
                custom_registry_folder.mkdir(parents=True, exist_ok=True)
                registry_file = custom_registry_folder.joinpath("registry.txt")
                urlretrieve(remote_registry, registry_file)  # noqa: S310
                return load_registry_from_file(registry_file)
        else:
            # If the branch is not the default version, check if the registry file exists
            custom_registry_folder = Path(str(ilr.files("ravenpy").joinpath("testing", branch)))
            custom_registry_folder.mkdir(parents=True, exist_ok=True)
            registry_file = custom_registry_folder.joinpath("registry.txt")
            with FileLock(custom_registry_folder.joinpath(".lock")):
                if not registry_file.exists():
                    urlretrieve(remote_registry, registry_file)  # noqa: S310
            return load_registry_from_file(registry_file)

    else:
        registry_file = Path(str(ilr.files("ravenpy").joinpath("testing/registry.txt")))

    if not registry_file.exists():
        msg = f"Registry file not found: {registry_file}"
        raise FileNotFoundError(msg)

    return load_registry_from_file(registry_file)


def yangtze(
    repo: str = TESTDATA_REPO_URL,
    branch: str = TESTDATA_BRANCH,
    cache_dir: str | Path = TESTDATA_CACHE_DIR,
    allow_updates: bool = True,
    force_download: bool = False,
):
    """
    Pooch registry instance for RavenPy test data.

    Parameters
    ----------
    repo : str
        URL of the repository to use when fetching testing datasets.
    branch : str
        Branch of repository to use when fetching testing datasets.
    cache_dir : str or Path
        The path to the directory where the data files are stored.
    allow_updates : bool
        If True, allow updates to the data files. Default is True.
    force_download : bool
        If True, force the download of the registry file even if it already exists.

    Returns
    -------
    pooch.Pooch
        The Pooch instance for accessing the RavenPy testing data.

    Notes
    -----
    There are three environment variables that can be used to control the behaviour of this registry:
        - ``RAVEN_TESTDATA_CACHE_DIR``: If this environment variable is set, it will be used as the
          base directory to store the data files.
          The directory should be an absolute path (i.e. it should start with ``/``).
          Otherwise, the default location will be used (based on ``platformdirs``, see :py:func:`pooch.os_cache`).
        - ``RAVEN_TESTDATA_REPO_URL``: If this environment variable is set, it will be used as the URL of
          the repository to use when fetching datasets. Otherwise, the default repository will be used.
        - ``RAVEN_TESTDATA_BRANCH``: If this environment variable is set, it will be used as the branch of
          the repository to use when fetching datasets. Otherwise, the default branch will be used.

    Examples
    --------
    Using the registry to download a file:

    .. code-block:: python

        import xarray as xr
        from ravenpy.testing import yangtze

        example_file = yangtze().fetch("example.nc")
        data = xr.open_dataset(example_file)
    """
    if pooch is None:
        raise ImportError(
            "The `pooch` package is required to fetch the RavenPy testing data. "
            "You can install it with `pip install pooch` or `pip install ravenpy[dev]`."
        )
    if not repo.endswith("/"):
        repo = f"{repo}/"
    remote = audit_url(urljoin(urljoin(repo, branch if branch.endswith("/") else f"{branch}/"), "data"))

    _yangtze = pooch.create(
        path=cache_dir,
        base_url=remote,
        version=default_testdata_version,
        version_dev=branch,
        allow_updates=allow_updates,
        registry=load_registry(branch=branch, repo=repo, force_download=force_download),
    )

    # Add a custom fetch method to the Pooch instance
    # Needed to address: https://github.com/readthedocs/readthedocs.org/issues/11763
    # Fix inspired by @bjlittle (https://github.com/bjlittle/geovista/pull/1202)
    _yangtze.fetch_diversion = _yangtze.fetch

    # Overload the fetch method to add user-agent headers
    @wraps(_yangtze.fetch_diversion)
    def _fetch(*args, **kwargs: bool | Callable) -> str:  # numpydoc ignore=GL08  # *args: str
        def _downloader(
            url: str,
            output_file: str | IO,
            poocher: pooch.Pooch,
            check_only: bool | None = False,
        ) -> None:
            """Download the file from the URL and save it to the save_path."""
            headers = {"User-Agent": f"RavenPy ({ravenpy.__version__})"}
            downloader = pooch.HTTPDownloader(headers=headers)
            return downloader(url, output_file, poocher, check_only=check_only)

        # default to our http/s downloader with user-agent headers
        kwargs.setdefault("downloader", _downloader)
        return _yangtze.fetch_diversion(*args, **kwargs)

    # Replace the fetch method with the custom fetch method
    _yangtze.fetch = _fetch

    return _yangtze


def get_file(
    name: str,
    _yangtze_kwargs: dict[str, Path | str | bool] | None = None,
):
    r"""
    Convenience function to get a file from the RavenPy testing data using the `yangtze` class.

    This is a thin wrapper around the `yangtze` class to make it easier to access RavenPy testing datasets.

    Parameters
    ----------
    name : str
        Name of the file containing the dataset.
    _yangtze_kwargs : dict
        Keyword arguments passed to the yangtze function.

    Returns
    -------
    str
        The path to the file.
    """
    if _yangtze_kwargs is None:
        _yangtze_kwargs = {}
    return yangtze(**_yangtze_kwargs).fetch(name)


def open_dataset(
    name: str,
    _yangtze_kwargs: dict[str, Path | str | bool] | None = None,
    **xr_kwargs: Any,
) -> Dataset:
    r"""
    Convenience function to open a dataset from the RavenPy testing data using the `yangtze` class.

    This is a thin wrapper around the `yangtze` class to make it easier to open RavenPy testing datasets.

    Parameters
    ----------
    name : str
        Name of the file containing the dataset.
    _yangtze_kwargs : dict
        Keyword arguments passed to the yangtze function.
    **xr_kwargs : Any
        Keyword arguments passed to xarray.open_dataset.

    Returns
    -------
    xarray.Dataset
        The dataset.

    See Also
    --------
    xarray.open_dataset : Open and read a dataset from a file or file-like object.
    yangtze : Pooch wrapper for accessing the RavenPy testing data.
    """
    if _yangtze_kwargs is None:
        _yangtze_kwargs = {}
    return _open_dataset(yangtze(**_yangtze_kwargs).fetch(name), **xr_kwargs)


def populate_testing_data(
    temp_folder: Path | None = None,
    repo: str = TESTDATA_REPO_URL,
    branch: str = TESTDATA_BRANCH,
    retry: int = 3,
    local_cache: Path = TESTDATA_CACHE_DIR,
) -> None:
    """
    Populate the local cache with the testing data.

    Parameters
    ----------
    temp_folder : Path, optional
        Path to a temporary folder to use as the local cache. If not provided, the default location will be used.
    repo : str, optional
        URL of the repository to use when fetching testing datasets.
    branch : str, optional
        Branch of raven-testdata to use when fetching testing datasets.
    retry : int
        Number of times to retry downloading the files in case of failure. Default: 3.
    local_cache : Path
        The path to the local cache. Defaults to the location set by the platformdirs library.
        The testing data will be downloaded to this local cache.
    """
    # Create the Pooch instance
    n = yangtze(repo=repo, branch=branch, cache_dir=temp_folder or local_cache)

    # Download the files
    errored_files = []
    for file in load_registry():
        msg = f"Downloading file `{file}` from remote repository..."
        logging.info(msg)
        for attempt in range(retry):
            try:
                n.fetch(file)
            except HTTPError:  # noqa: PERF203
                msg = f"Failed to download file `{file}` on attempt {attempt + 1}."
                logging.info(msg)
            else:
                logging.info("File was downloaded successfully.")
                break
        else:
            msg = f"Failed to download file `{file}` after {retry} attempts."
            logging.error(msg)
            errored_files.append(file)

    if errored_files:
        msg = f"The following files were unable to be downloaded: {errored_files}"
        logging.error(msg)


def gather_testing_data(
    worker_cache_dir: str | os.PathLike[str] | Path,
    worker_id: str,
    _cache_dir: str | os.PathLike[str] | None = TESTDATA_CACHE_DIR,
) -> None:
    """
    Gather testing data across workers.

    Parameters
    ----------
    worker_cache_dir : str or Path
        The directory to store the testing data.
    worker_id : str
        The worker ID.
    _cache_dir : str or Path, optional
        The directory to store the testing data. Default is None.

    Raises
    ------
    ValueError
        If the cache directory is not set.
    FileNotFoundError
        If the testing data is not found.
    """
    if _cache_dir is None:
        raise ValueError(
            "The cache directory must be set. Please set the `cache_dir` parameter or the `RAVEN_TESTDATA_CACHE_DIR` environment variable."
        )
    cache_dir = Path(_cache_dir)

    if worker_id == "master":
        populate_testing_data(branch=TESTDATA_BRANCH)
    else:
        cache_dir.mkdir(exist_ok=True, parents=True)
        lockfile = cache_dir.joinpath(".lock")
        test_data_being_written = FileLock(lockfile)
        with test_data_being_written:
            # This flag prevents multiple calls from re-attempting to download testing data in the same pytest run
            populate_testing_data(branch=TESTDATA_BRANCH)
            cache_dir.joinpath(".data_written").touch()
        with test_data_being_written.acquire():
            if lockfile.exists():
                lockfile.unlink()
        copytree(cache_dir.joinpath(default_testdata_version), worker_cache_dir)


# Testing Utilities ###


def audit_url(url: str, context: str | None = None) -> str:
    """
    Check if the URL is well-formed.

    Parameters
    ----------
    url : str
        The URL to check.
    context : str, optional
        Additional context to include in the error message. Default is None.

    Returns
    -------
    str
        The URL if it is well-formed.

    Raises
    ------
    URLError
        If the URL is not well-formed.
    """
    msg = ""
    result = urlparse(url)
    if result.scheme == "http":
        msg = f"{context if context else ''} URL is not using secure HTTP: '{url}'".strip()
    if not all([result.scheme, result.netloc]):
        msg = f"{context if context else ''} URL is not well-formed: '{url}'".strip()

    if msg:
        LOGGER.error(msg)
        raise URLError(msg)
    return url



import datetime as dt
import pathlib
from typing import Union

import cftime
import pytest
from pydantic import ConfigDict, Field, ValidationError
from pydantic.dataclasses import dataclass
from pymbolic.primitives import Variable

from ravenpy.config import commands as rc
from ravenpy.config import options
from ravenpy.config.rvs import RV, RVI, Config, optfield


def test_optfield():
    class Test(RV):
        a: bool = optfield(alias="a")

    t = Test()
    assert not t.__class__.model_fields["a"].is_required()


def test_rvi_datetime():
    exp = cftime.datetime(1990, 1, 1, calendar="PROLEPTIC_GREGORIAN")

    rvi = RVI(start_date=dt.datetime(1990, 1, 1))
    assert rvi.start_date == exp

    rvi = RVI(start_date="1990-01-01")
    assert rvi.start_date == exp

    rvi = RVI(start_date="1990-01-01T00:00:00")
    assert rvi.start_date == exp

    rvi = RVI(end_date=cftime.datetime(1990, 1, 1))
    assert rvi.end_date == exp

    rvi = RVI(start_date=dt.datetime(1990, 1, 1), calendar="NOLEAP")
    assert rvi.start_date == cftime.datetime(1990, 1, 1, calendar="NOLEAP")

    with pytest.raises(ValidationError):
        rvi = RVI(start_date=(dt.datetime(1990, 1, 1),))


def test_duplicate_simple():
    conf = Config(start_date="1990-01-01")

    # Updating values with an alias and an attribute name
    out = conf.duplicate(Duration=10, debug_mode=True)

    assert isinstance(out.start_date, cftime.datetime)
    assert out.duration == 10
    assert out.debug_mode


def test_duplicate_emulator(gr4jcn_config):
    conf, params = gr4jcn_config
    conf.duplicate()

    conf.duplicate(params=params)


def test_set_params():
    @dataclass(config=ConfigDict(arbitrary_types_allowed=True))
    class P:
        X01: Union[Variable, float] = Variable("X01")

    class MySymbolicEmulator(Config):
        params: P = P()
        rain_snow_transition: rc.RainSnowTransition = Field(
            default=rc.RainSnowTransition(temp=P.X01, delta=2),
            alias="RainSnowTransition",
        )

    # Assignment through set_params -> new instance
    s = MySymbolicEmulator()
    num = s.set_params([0.5])
    assert num.rain_snow_transition.temp == 0.5

    s = MySymbolicEmulator()
    num = s.set_params({"X01": 0.5})
    assert num.rain_snow_transition.temp == 0.5


def test_solution(yangtze):
    sol = pathlib.Path(yangtze.fetch("gr4j_cemaneige/solution.rvc"))
    conf = Config().set_solution(sol)
    assert len(conf.hru_state_variable_table) == 1
    assert conf.hru_state_variable_table[0].data["ATMOSPHERE"] == 821.98274
    assert conf.hru_state_variable_table[0].data["ATMOS_PRECIP"] == -1233.16

    assert len(conf.basin_state_variables) == 1
    assert conf.basin_state_variables[0].channel_storage == 0
    assert conf.basin_state_variables[0].qout == (1, 13.21660, 13.29232)

    assert ":BasinIndex 1 watershed" in conf.rvc


def test_rvh_from_extractor(yangtze):
    from ravenpy.extractors import BasinMakerExtractor, open_shapefile

    shp = yangtze.fetch("basinmaker/drainage_region_0175_v2-1/finalcat_info_v2-1.zip")
    bm = BasinMakerExtractor(open_shapefile(shp))

    # Smoke test
    Config(**bm.extract(hru_from_sb=True))


def test_config(dummy_config):
    cls, P = dummy_config
    conf = cls(Calendar="NOLEAP")

    # Set params
    num = conf.set_params([0.5])
    assert num.air_snow_coeff == 0.5

    with pytest.raises(ValueError):
        num.set_params([0.6])

    # Instantiate with numerical params
    assert conf.model_config["populate_by_name"]
    # nt = cls(params=[0.5], Calendar="NOLEAP")
    # assert nt.air_snow_coeff == 0.5


def test_custom_subclass(dummy_config, tmp_path):
    """Test that users can subclass RV and Config."""
    cls, P = dummy_config

    # Custom RVI
    class myRVI(RVI):
        run_name: str = Field("myRunName", alias="RunName")

    # Custom config with custom RVI
    class MyConfig(myRVI, cls):
        params: P = P()
        enkf_mode: options.EnKFMode = optfield(alias="EnKFMode")

    # Make sure rv files can be written
    conf = MyConfig(EnKFMode="ENKF_SPINUP").set_params([0.5])
    conf.write_rv(workdir=tmp_path)
    assert conf.run_name == "myRunName"
    assert "myRunName" in conf._rv("RVI")
    assert "EnKFMode" not in conf._rv("RVI")


def test_hru_filter():
    """Test that unrecognized HRU types are filtered out."""
    from ravenpy.config.emulators.gr4jcn import LakeHRU
    from ravenpy.config.emulators.hbvec import HRUs, LandHRU

    with pytest.warns(UserWarning):
        hrus = HRUs([LandHRU(), LandHRU(), LakeHRU()])

    # The GR4J lake HRU is not part of the HBVEC config.
    assert len(hrus.root) == 2



import datetime as dt
import urllib.error
from pathlib import Path

import numpy as np
import pytest
import xarray

from ravenpy.extractors.forecasts import get_CASPAR_dataset, get_ECCC_dataset
from ravenpy.testing.utils import (
    default_testdata_cache,
    default_testdata_version,
    open_dataset,
    yangtze,
)


@pytest.mark.online
class TestGet:
    def test_get_CASPAR_dataset(self):
        ds, _ = get_CASPAR_dataset("GEPS", dt.datetime(2018, 8, 31))

    @pytest.mark.xfail(error=OSError, reason="Network may be unreliable", strict=False)
    def test_get_ECCC_dataset(self):
        ds, _ = get_ECCC_dataset("GEPS")

        ns = np.datetime64("now") - ds.time.isel(time=0).values
        n_hours = ns / np.timedelta64(1, "h")

        assert n_hours <= 36


@pytest.mark.online
class TestRemoteFileAccess:
    dap_url = "http://test.opendap.org:80/opendap/data/nc/"
    git_url = "https://github.com/Ouranosinc/raven-testdata"
    branch = "main"

    @pytest.mark.xfail(
        raises=urllib.error.URLError,
        reason="Get file is API rate limited",
        strict=False,
    )
    def test_get_file_default_cache(self):
        file = yangtze(branch=self.branch).fetch(fname="ostrich-hbvec/raven-hbvec-salmon.rvi")

        assert Path(default_testdata_cache).exists()
        assert Path(file).is_file()
        with Path(file).open() as f:
            header = f.read()
            assert ":FileType          rvi ASCII Raven 2.8.2" in header

    def test_open_dataset(
        self,
        tmp_path,
    ):
        cache_dir = tmp_path / "yangtze_cache"
        ds = open_dataset(
            name="raven-gr4j-cemaneige/Salmon-River-Near-Prince-George_meteo_daily.nc",
            _yangtze_kwargs={
                "branch": self.branch,
                "cache_dir": cache_dir,
                "force_download": True,
            },
        )

        assert (
            Path(cache_dir)
            .joinpath(
                default_testdata_version,
                "raven-gr4j-cemaneige",
                "Salmon-River-Near-Prince-George_meteo_daily.nc",
            )
            .exists()
        )
        assert isinstance(ds, xarray.Dataset)
